"""Wherein is contained the concrete SQLite implementation of the Index."""
# Pylint doesn't like the similarity between this file and the SQLite file, but
# it doesn't make sense to write shared functions for them. So ignore pylint.
# pylint: disable=duplicate-code
import os
import warnings
from datetime import datetime

import ast
import pandas as pd
# Try-Except required to make psycopg2/sqlalchemy an optional dependency.
try:
    # pylint: disable=unused-import
    import importlib
    assert importlib.util.find_spec('psycopg2')
    import sqlalchemy as sqla # noqa: F401
    # pylint: enable=unused-import
except ImportError:
    _HAS_REQUIRED_DEPS = False
except AssertionError:
    _HAS_REQUIRED_DEPS = False
else:
    _HAS_REQUIRED_DEPS = True

from .index_abc import IndexABC
from .create_index import create_index_from_fs

class IndexSQL(IndexABC):
    """Concrete implementation of Index, using SQL."""
    def __init__(self, file_system, pantry_path, **kwargs):
        """Initializes the Index class.

        Parameters
        ----------
        file_system: fsspec object
            The fsspec object which hosts the pantry we desire to index.
        pantry_path: str
            Path to the pantry root which we want to index.
        **database_name: str (default='weave_db')
            DB to be used. If none is set, the environment variable
            WEAVE_SQL_DB_NAME is check. If not set, default to weave_db.
        **pantry_schema: str (default=<pantry_path>)
            The schema to use for the pantry. If none is set, defaults to the
            pantry path (with _ replacements when necessary).
        """
        if not _HAS_REQUIRED_DEPS:
            raise ImportError("Missing Dependencies. The packages: 'psycopg2'"
                              "'sqlalchemy' are required to use this class")

        # Check that the required environment variables are set.
        try:
            # We want to fail early if these are not set, so we check them here
            # even though they are not used until we connect.
            # Pylint thinks this is pointless, so pylint is getting ignored.
            # pylint: disable=pointless-statement
            self._sql_connection = {
                'host': os.environ["WEAVE_SQL_HOST"],
                'username': os.environ["WEAVE_SQL_USERNAME"],
                'password': os.environ["WEAVE_SQL_PASSWORD"],
                'port': int(os.environ.get("WEAVE_SQL_PORT", 5432)),
            }
            # pylint: enable=pointless-statement
        except KeyError as key_error:
            raise KeyError("The following environment variables must be set "
                           "to use this class: WEAVE_SQL_HOST, "
                           "WEAVE_SQL_USERNAME, WEAVE_SQL_PASSWORD."
                          ) from key_error

        self._file_system = file_system
        self._pantry_path = pantry_path

        # Set the database name (defaults to weave_db). DATABASE MUST ALREADY
        # EXIST. If it does not, the user must create it manually.
        self._database_name = kwargs.get(
            "database_name",
            os.environ.get("WEAVE_SQL_DB_NAME", "weave_db"),
        )
        self._database_name = self._database_name.replace("-", "_")

        # Set the schema name (defaults to pantry_path). If the schema does not
        # exist, it will be created.
        d_schema_name = self._pantry_path.replace(os.sep, "_")\
            .replace("-", "_")
        if d_schema_name == "":
            d_schema_name = "weave"
        self._pantry_schema = kwargs.get("pantry_schema", d_schema_name)
        self._pantry_schema = self._pantry_schema.lower()

        self._engine = sqla.create_engine(sqla.engine.url.URL(
            drivername="postgresql",
            username=self._sql_connection['username'],
            password=self._sql_connection['password'],
            host=self._sql_connection['host'],
            database=self.database_name,
            query={},
            port=self._sql_connection['port'],
        ))

        self._create_schema()
        self._create_tables()

    # Create read-only properties for the database and schema names.
    @property
    def database_name(self):
        """The database name of the index."""
        return self._database_name

    @property
    def pantry_schema(self):
        """The schema name of the index."""
        return self._pantry_schema

    def execute_sql(self, sql_query, params=None, commit=False):
        """Executes the given SQL query. Returns the results.

        Parameters
        ----------
        sql_query: str or sqlalchemy.sql.text
            The SQL query to be executed.
        params: dict (optional)
            The parameters to be used in the query.
        commit: bool (default=False)
            Whether or not to commit the query.

        Returns
        ----------
        (list, list) or int or None
            Returns the results of the query and the column names as a tuple.
            If statement affects rows, returns the number of rows affected.
            If the query does not return any results, returns None.
        """
        try:
            with self._engine.connect() as connection:
                if isinstance(sql_query, sqla.sql.elements.TextClause):
                    query = sql_query
                elif isinstance(sql_query, str):
                    query = sqla.sql.text(sql_query)
                else:
                    raise ValueError(
                        "sql_query should be a str or a "
                        "sqlalchemy TextClause object"
                    )

                if params is not None:
                    if not isinstance(params, dict):
                        raise TypeError("params should be a dict")

                    # Execute the SQL query with parameters
                    result = connection.execute(query, params)
                else:
                    # Execute the SQL query without parameters
                    result = connection.execute(query)

                if commit:
                    connection.commit()

                # Fetch and return the results
                if result.returns_rows:
                    return result.fetchall(), list(result.keys())

                # Return rows affected (used for INSERT, DELETE, etc.)
                if result.rowcount != -1:
                    return result.rowcount

                return None

        except sqla.exc.SQLAlchemyError as err:
            raise err

    def _create_schema(self):
        """Create the schema if it does not already exist."""
        with self._engine.connect() as connection:
            if not connection.dialect.has_schema(connection,
                                                 self.pantry_schema):
                self.execute_sql(f"CREATE SCHEMA {self.pantry_schema};",
                                commit=True)

    def _create_tables(self):
        """Create the required tables if they do not already exist."""
        # THIS NEEDS TO BE UPDATED MANUALLY IF NEW COLUMNS ARE ADDED TO INDEX.
        # THE INSERT IN OTHER FUNCTIONS USE config.index_schema(), BUT THAT
        # CAN'T BE USED HERE AS TYPE NEEDS TO BE SPECIFIED.
        if not sqla.inspect(self._engine).has_table("pantry_index",
                                                    self.pantry_schema):
            self.execute_sql(
                f"""
                CREATE TABLE {self.pantry_schema}.pantry_index (
                    uuid varchar(64),
                    upload_time INT,
                    parent_uuids TEXT,
                    basket_type TEXT,
                    label TEXT,
                    weave_version TEXT,
                    address TEXT,
                    storage_type TEXT,
                    PRIMARY KEY(uuid),
                    UNIQUE(uuid)
                );
                """, commit=True
            )
        if not sqla.inspect(self._engine).has_table("parent_uuids",
                                                    self.pantry_schema):
            self.execute_sql(
                f"""
                CREATE TABLE {self.pantry_schema}.parent_uuids (
                    uuid varchar(64),
                    parent_uuid varchar(64),
                    PRIMARY KEY(uuid, parent_uuid),
                    UNIQUE(uuid, parent_uuid)
                );
                """, commit=True
            )

    @property
    def file_system(self):
        """The file system of the pantry referenced by this Index."""
        return self._file_system

    @property
    def pantry_path(self):
        """The pantry path referenced by this Index."""
        return self._pantry_path

    def generate_metadata(self, **kwargs):
        """(Deprecated) Generate the metadata for the index."""
        warnings.warn(
            UserWarning("This function is being deprecated in a future weave "
                        "release. Please use generate_config() instead.")
        )
        return self.generate_config(**kwargs)

    def generate_config(self, **kwargs):
        """Populates the metadata for the index.

        Parameters
        ----------
        **kwargs unused for this function.

        Returns
        ----------
        dict
            Returns a dictionary of the metadata.
        """
        return {
            "database_host": self._sql_connection['host'],
            "database_port": self._sql_connection['port'],
            "database_name": self.database_name,
            "database_schema": self.pantry_schema,
        }

    def generate_index(self, **kwargs):
        """Populates the index from the file system.

        Generate the index by scraping the pantry and adding the manifest data
        of found baskets to the index.

        Parameters
        ----------
        **kwargs unused for this function.
        """
        if not isinstance(self.pantry_path, str):
            raise TypeError("'pantry_path' must be a string: "
                            f"'{self.pantry_path}'")

        if not self.file_system.exists(self.pantry_path):
            raise FileNotFoundError("'pantry_path' does not exist: "
                                    f"'{self.pantry_path}'")

        basket_jsons = [x for x in self.file_system.find(self.pantry_path)
            if x.endswith("basket_manifest.json")
        ]

        for basket_json_address in basket_jsons:
            entry = create_index_from_fs(basket_json_address,
                                         file_system=self.file_system)
            if not entry.empty:
                if len(self.get_rows(entry['uuid'].iloc[0])) == 0:
                    self.track_basket(entry)

    def clear_index(self, refresh=False, **kwargs):
        """Deletes/clears the previously populated index.

        Deletes the index entirely (by dropping and re-creating the schema)
        optionally repopulating the new one if the refresh flag is set to True.

        Parameters
        ----------
        refresh: bool (default=False)
            Regenerates the index after clearing the old one.

        **kwargs unused for this function.
        """
        # Drop this index's schema, and then re-create the schema and tables.
        self.drop_index()
        self._create_schema()
        self._create_tables()

        # Optionally re-populate the tables.
        if refresh:
            self.generate_index()

    def drop_index(self):
        """Drops all data stored in the index backend and drops the schema.

        The schemas and tables are NOT regenerated in this call.
        Use clear_index if you wish to have them rebuilt with nothing in them.
        """
        # Drop this index's schema, and then re-create the schema and tables.
        self.execute_sql(
            f"DROP SCHEMA IF EXISTS {self.pantry_schema} CASCADE;",
            commit=True,
        )

    def to_pandas_df(self, max_rows=None, offset=0, **kwargs):
        """Returns the pandas dataframe representation of the index.

        Parameters
        ----------
        max_rows: int or None (default=None)
            Max rows returned in the pandas dataframe. If None, all rows will
            be returned.
        offset: int (default=0)
            Offset from the beginning of the index to begin the query
        **kwargs unused for this function.

        Returns
        ----------
        pandas.DataFrame
            Returns a dataframe of the manifest data of the baskets in the
            pantry.
        """
        # Get the rows from the index as a list of lists, then get the columns.
        if max_rows is None:
            result, columns = self.execute_sql(
                 f"""SELECT *
                     FROM {self.pantry_schema}.pantry_index
                     ORDER BY UUID""")
        else:
            result, columns = self.execute_sql(
             f"""SELECT *
                 FROM {self.pantry_schema}.pantry_index
                 ORDER BY UUID
                 OFFSET (:offset) ROWS
                 FETCH FIRST (:max_rows) ROWS ONLY""",
            {"offset": offset, "max_rows": max_rows},
        )

        result = [list(row) for row in result]

        ind_df = pd.DataFrame(result, columns=columns)
        ind_df["parent_uuids"] = ind_df["parent_uuids"].apply(ast.literal_eval)
        ind_df["upload_time"] = pd.to_datetime(
            ind_df["upload_time"],
            unit="s",
            origin="unix",
        )
        return ind_df

    def track_basket(self, entry_df, **kwargs):
        """Track a basket (or many baskets) from the pantry with the Index.

        Parameters
        ----------
        entry_df: pd.DataFrame
            Uploaded baskets' manifest data to append to the index.

        **kwargs unused for this function.
        """
        # Insert the parent_uuids.
        sql = sqla.text(
            f"INSERT INTO {self.pantry_schema}.parent_uuids "
            "(uuid, parent_uuid) "
            "SELECT :uuid, :parent_uuid "
            "WHERE NOT EXISTS "
            f"(SELECT 1 FROM {self.pantry_schema}.parent_uuids "
            "WHERE uuid = CAST(:uuid AS text) "
            "AND parent_uuid = CAST(:parent_uuid AS text));"
        )

        # Loop all uuids and parent uuids (list of lists).
        for _, entry in entry_df[["uuid", "parent_uuids"]].iterrows():
            # Loop all parent uuids (now a list of strings)
            for parent_uuid in entry.parent_uuids:
                self.execute_sql(
                    sql,
                    {
                        "uuid": entry.uuid,
                        "parent_uuid": parent_uuid,
                    },
                    commit=True,
                )

        # Convert the parent_uuids to a string, and the upload_time to an int.
        entry_df.loc[:,"parent_uuids"] = entry_df["parent_uuids"].astype(str)
        entry_df["upload_time"] = (
            entry_df["upload_time"].astype(int) // 1e9
        ).astype(int)

        for basket_dict in entry_df.to_dict(orient="records"):
            index_columns = list(basket_dict.keys())
            # Insert into pantry_index.
            sql = sqla.text(
                f"INSERT INTO {self.pantry_schema}.pantry_index ("
                f"{', '.join([f'{column}' for column in index_columns])}) "
                "SELECT "
                f"{', '.join([f':{column}' for column in index_columns])} "
                "WHERE NOT EXISTS "
                f"(SELECT 1 FROM {self.pantry_schema}.pantry_index "
                "WHERE uuid = CAST(:uuid AS text));"
            )
            self.execute_sql(sql, basket_dict, commit=True)

    def untrack_basket(self, basket_address, **kwargs):
        """Remove a basket from being tracked of given UUID or path.

        Parameters
        ----------
        basket_address: str or [str]
            Argument can take one of two forms: either a path to the basket
            directory, or the UUID of the basket. These may also be passed in
            as a list.

        **kwargs unused for this function.
        """
        if not isinstance(basket_address, list):
            basket_address = [basket_address]

        if self.file_system.exists(os.fspath(basket_address[0])):
            uuids, _ = self.execute_sql(
                sqla.text(
                    f"SELECT uuid FROM {self.pantry_schema}.pantry_index "
                    f"WHERE address IN (:basket_address)"
                ),
                {"basket_address": ','.join(basket_address)}
            )
            uuids = [uuid[0] for uuid in uuids]
        else:
            uuids = basket_address

        # Delete from pantry_index.
        query = sqla.text(
            f"DELETE FROM {self.pantry_schema}.pantry_index "
            " WHERE uuid IN ( "
                " SELECT unnest(CAST(:uuids AS text[]))"
            ");"
        )
        rowcount = self.execute_sql(
            query,
            {"uuids": uuids},
            commit=True
        )

        if rowcount != len(uuids):
            warnings.warn(
                UserWarning(
                    "Incomplete Request. Index could not untrack baskets, "
                    "as some were not being tracked to begin with.",
                    len(uuids) - rowcount
                )
            )

        # Delete from parent_uuids.
        query = (
            f"DELETE FROM {self.pantry_schema}.parent_uuids "
            " WHERE uuid IN ( "
                " SELECT unnest(CAST(:uuids AS text[]))"
            ");"
        )
        self.execute_sql(query, {"uuids": uuids}, commit=True)

    def get_rows(self, basket_address, **kwargs):
        """Returns a pd.DataFrame row information of given UUID or path.

        Parameters
        ----------
        basket_address: str or [str]
            Argument can take one of two forms: either a path to the basket
            directory, or the UUID of the basket. These may also be passed in
            as a list.
        **kwargs unused for this function.

        Returns
        ----------
        pandas.DataFrame
            Manifest information for the requested basket(s).
        """
        if not isinstance(basket_address, list):
            basket_address = [basket_address]

        if self.file_system.exists(os.fspath(basket_address[0])):
            id_column = "address"
        else:
            id_column = "uuid"

        baskets = [f"'{e}'" for e in basket_address]
        query = (
            f"SELECT * FROM {self.pantry_schema}.pantry_index "
            f"WHERE {id_column} IN ({','.join(baskets)});"
        )
        results, columns = self.execute_sql(query,)
        results = [list(row) for row in results]

        ind_df = pd.DataFrame(
            results,
            columns=columns
        )
        ind_df["parent_uuids"] = ind_df["parent_uuids"].apply(ast.literal_eval)
        ind_df["upload_time"] = pd.to_datetime(
            ind_df["upload_time"],
            unit="s",
            origin="unix",
        )
        return ind_df

    def get_parents(self, basket_address, **kwargs):
        """Returns a pandas dataframe of all parents of a basket.

        Parameters
        ----------
        basket_address: str
            Argument can take one of two forms: either a path to the basket
            directory, or the UUID of the basket.
        **max_gen_level: int (optional)
            This indicates the maximum generation level that will be reported.

        Returns
        ----------
        pandas.DataFrame containing all the manifest data AND generation level
        of parents (and recursively their parents) of the given basket.
        """

        max_gen_level = kwargs.get("max_gen_level", 999)

        if self.file_system.exists(os.fspath(basket_address)):
            id_column = "address"
        else:
            id_column = "uuid"

        basket_uuid, _ = self.execute_sql(
            f"SELECT uuid FROM {self.pantry_schema}.pantry_index "
            f"WHERE {id_column} = :basket_address",
            {"basket_address": basket_address, "id_column": id_column}
        )

        if basket_uuid is None or len(basket_uuid) == 0:
            raise FileNotFoundError(
                f"basket path or uuid does not exist '{basket_address}'"
            )
        basket_uuid = basket_uuid[0][0]

        results, columns = self.execute_sql(
            f"""
            WITH RECURSIVE child_record AS (
                SELECT
                    0 AS level,
                    CAST(:basket_uuid AS varchar) AS id,
                    CAST(:basket_uuid AS varchar) AS path
                UNION ALL
                SELECT
                    child_record.level + 1,
                    CAST(parent_uuids.parent_uuid AS varchar) AS id,
                    CAST(child_record.path || '/' || parent_uuids.parent_uuid
                        AS varchar) AS path
                FROM {self.pantry_schema}.parent_uuids
                JOIN child_record ON parent_uuids.uuid = child_record.id
                WHERE
                    path NOT LIKE CONCAT(parent_uuids.parent_uuid, '/%')
                    AND path NOT LIKE CONCAT('%', parent_uuids.parent_uuid)
                    AND path NOT LIKE
                        CONCAT('%', parent_uuids.parent_uuid, '/%')
                AND child_record.level < :max_gen_level
            )
            SELECT pantry_index.*, child_record.level, child_record.path
            FROM {self.pantry_schema}.pantry_index as pantry_index
            JOIN child_record ON pantry_index.uuid = child_record.id
            ORDER BY child_record.level ASC;
            """,
            {"basket_uuid": basket_uuid, "max_gen_level": max_gen_level}
        )

        results = [list(row) for row in results]
        columns[columns.index("level")] = "generation_level"

        parent_df = pd.DataFrame(results, columns=columns)
        parent_df = parent_df.drop_duplicates()
        parent_df = parent_df[parent_df["uuid"] != basket_uuid]

        if parent_df.empty:
            return parent_df
        parent_df["parent_uuids"] = parent_df["parent_uuids"].apply(
            ast.literal_eval
        )
        parent_df["upload_time"] = pd.to_datetime(
            parent_df["upload_time"],
            unit="s",
            origin="unix",
        )

        for _, row in parent_df.iterrows():
            for prev in row['path'].split('/'):
                if prev in row["parent_uuids"]:
                    raise ValueError(
                        f"Parent-Child loop found at uuid: {basket_uuid}"
                    )

        parent_df.drop(columns="path", inplace=True)

        return parent_df

    def get_children(self, basket_address, **kwargs):
        """Returns a pandas dataframe of all children of a basket.

        Parameters
        ----------
        basket_address: str
            Argument can take one of two forms: either a path to the basket
            directory, or the UUID of the basket.
        **min_gen_level: int (optional)
            This indicates the minimum generation level that will be reported.

        Returns
        ----------
        pandas.DataFrame containing all the manifest data AND generation level
        of children (and recursively their children) of the given basket.
        """

        min_gen_level = kwargs.get("min_gen_level", -999)

        if self.file_system.exists(os.fspath(basket_address)):
            id_column = "address"
        else:
            id_column = "uuid"

        basket_uuid, _ = self.execute_sql(
            f"SELECT uuid FROM {self.pantry_schema}.pantry_index "
            f"WHERE {id_column} = :basket_address",
            {"basket_address": basket_address, "id_column": id_column}
        )

        if basket_uuid is None or len(basket_uuid) == 0:
            raise FileNotFoundError(
                f"basket path or uuid does not exist '{basket_address}'"
            )
        basket_uuid = basket_uuid[0][0]

        results, columns = self.execute_sql(
            f"""
            WITH RECURSIVE child_record AS (
                SELECT
                    0 AS level,
                    CAST(:basket_uuid AS varchar) AS id,
                    CAST(:basket_uuid AS varchar) AS path
                UNION ALL
                SELECT 
                    child_record.level - 1,
                    CAST(parent_uuids.uuid AS varchar) AS id,
                    CAST(child_record.path || '/' || parent_uuids.uuid
                        AS varchar) AS path
                FROM {self.pantry_schema}.parent_uuids
                JOIN child_record ON parent_uuids.parent_uuid = child_record.id
                WHERE 
                    path NOT LIKE CONCAT(parent_uuids.uuid, '/%')
                    AND path NOT LIKE CONCAT('%', parent_uuids.uuid)
                    AND path NOT LIKE CONCAT('%', parent_uuids.uuid, '/%')
                AND child_record.level > :min_gen_level
            )
            SELECT pantry_index.*, child_record.level, child_record.path
            FROM {self.pantry_schema}.pantry_index as pantry_index
            JOIN child_record ON pantry_index.uuid = child_record.id
            ORDER BY child_record.level DESC;
            """,
            {"basket_uuid": basket_uuid, "min_gen_level": min_gen_level}
        )

        results = [list(row) for row in results]
        columns[columns.index("level")] = "generation_level"

        child_df = pd.DataFrame(results, columns=columns)
        child_df = child_df.drop_duplicates()
        child_df["parent_uuids"] = child_df["parent_uuids"].apply(
            ast.literal_eval
        )
        child_df["upload_time"] = pd.to_datetime(
            child_df["upload_time"],
            unit="s",
            origin="unix",
        )

        parents = {}
        for _, row in child_df.iterrows():
            parents[row['uuid']] = row["parent_uuids"]
            for prev in row['path'].split('/'):
                if row['uuid'] in parents[prev]:
                    raise ValueError(
                        f"Parent-Child loop found at uuid: {basket_uuid}"
                    )

        child_df = child_df[child_df['uuid'] != basket_uuid]
        child_df.drop(columns="path", inplace=True)

        return child_df

    def get_baskets_of_type(self, basket_type, max_rows=None,
                            offset=0, **kwargs):
        """Returns a pandas dataframe containing baskets of basket_type.

        Parameters
        ----------
        basket_type: str
            The basket type to filter for.
        max_rows: int or None (default=None)
            Max rows returned in the pandas dataframe. If None, all rows will
            be returned.
        offset: int (default=0)
            Offset from the beginning of the index to begin the query

        **kwargs unused for this function.

        Returns
        ----------
        pandas.DataFrame containing the manifest data of baskets of the type.
        """
        if max_rows is None:
            result, columns = self.execute_sql(
                f"""SELECT *
                    FROM {self.pantry_schema}.pantry_index
                    WHERE basket_type = :basket_type
                    ORDER BY UUID""",
                {"basket_type": basket_type},
            )
        else:
            result, columns = self.execute_sql(
                f"""SELECT *
                    FROM {self.pantry_schema}.pantry_index
                    WHERE basket_type = :basket_type
                    ORDER BY UUID
                    OFFSET (:offset) ROWS
                    FETCH FIRST (:max_rows) ROWS ONLY""",
                {"offset": offset, "max_rows": max_rows,
                 "basket_type": basket_type},
            )
        result = [list(row) for row in result]

        ind_df = pd.DataFrame(result, columns=columns)
        ind_df["parent_uuids"] = ind_df["parent_uuids"].apply(ast.literal_eval)
        ind_df["upload_time"] = pd.to_datetime(
            ind_df["upload_time"],
            unit="s",
            origin="unix",
        )
        return ind_df

    def get_baskets_of_label(self, basket_label, max_rows=None,
                             offset=0, **kwargs):
        """Returns a pandas dataframe containing baskets with label.

        Parameters
        ----------
        basket_label: str
            The label to filter for.
        max_rows: int or None (default=None)
            Max rows returned in the pandas dataframe. If None, all rows will
            be returned.
        offset: int (default=0)
            Offset from the beginning of the index to begin the query

        **kwargs unused for this function.

        Returns
        ----------
        pandas.DataFrame containing the manifest data of baskets with the label
        """
        if max_rows is None:
            result, columns = self.execute_sql(
                f"""SELECT *
                    FROM {self.pantry_schema}.pantry_index
                    WHERE label = :basket_label
                    ORDER BY UUID""",
                {"basket_label": basket_label},
            )
        else:
            result, columns = self.execute_sql(
                f"""SELECT *
                    FROM {self.pantry_schema}.pantry_index
                    WHERE label = :basket_label
                    ORDER BY UUID
                    OFFSET (:offset) ROWS
                    FETCH FIRST (:max_rows) ROWS ONLY""",
                {"offset": offset, "max_rows": max_rows,
                 "basket_label": basket_label},
            )
        result = [list(row) for row in result]

        ind_df = pd.DataFrame(result, columns=columns)
        ind_df["parent_uuids"] = ind_df["parent_uuids"].apply(ast.literal_eval)
        ind_df["upload_time"] = pd.to_datetime(
            ind_df["upload_time"],
            unit="s",
            origin="unix",
        )
        return ind_df

    def get_baskets_by_upload_time(self, start_time=None, end_time=None,
                                   max_rows=None, offset=0, **kwargs):
        """Returns a pandas dataframe of baskets uploaded between two times.

        Parameters
        ----------
        start_time: datetime.datetime (optional)
            The start datetime object to filter between. If None, will filter
            from the beginning of time.
        end_time: datetime.datetime (optional)
            The end datetime object to filter between. If None, will filter
            to the current datetime.
        max_rows: int or None (default=None)
            Max rows returned in the pandas dataframe. If None, all rows will
            be returned.
        offset: int (default=0)
            Offset from the beginning of the index to begin the query

        **kwargs unused for this function.

        Returns
        ----------
        pandas.DataFrame containing the manifest data of baskets uploaded
        between the start and end times.
        """
        super().get_baskets_by_upload_time(start_time, end_time)
        if start_time is None and end_time is None:
            return self.to_pandas_df(max_rows=max_rows, offset=offset)

        pre_query = f"SELECT * FROM {self.pantry_schema}.pantry_index "
        post_query = """ORDER BY UUID """
        limit_query = """OFFSET (:offset) ROWS
                        FETCH FIRST (:max_rows) ROWS ONLY"""

        columns = None
        if start_time and end_time:
            start_time = int(datetime.timestamp(start_time))
            end_time = int(datetime.timestamp(end_time))
            query = "WHERE upload_time >= :start_time " \
                    "AND upload_time <= :end_time "
            results, columns = self.execute_sql(
                pre_query + query + post_query +
                    (limit_query if max_rows else ""),
                {"offset": offset,
                 "max_rows": max_rows,
                 "start_time": start_time,
                 "end_time": end_time
                })
        elif start_time:
            start_time = int(datetime.timestamp(start_time))
            query = "WHERE upload_time >= :start_time "
            results, columns = self.execute_sql(
                pre_query + query + post_query +
                    (limit_query if max_rows else ""),
                {"offset": offset,
                 "max_rows": max_rows,
                 "start_time": start_time,
                })
        elif end_time:
            end_time = int(datetime.timestamp(end_time))
            query = "WHERE upload_time <= :end_time "
            results, columns = self.execute_sql(
                pre_query + query + post_query +
                    (limit_query if max_rows else ""),
                {"offset": offset,
                 "max_rows": max_rows,
                 "end_time": end_time,
                })

        results = [list(row) for row in results]

        ind_df = pd.DataFrame(results, columns=columns)
        ind_df["parent_uuids"] = ind_df["parent_uuids"].apply(ast.literal_eval)
        ind_df["upload_time"] = pd.to_datetime(
            ind_df["upload_time"],
            unit="s",
            origin="unix",
        )
        return ind_df

    def query(self, expr, **kwargs):
        """Returns a pandas dataframe of the results of the expression.

        Parameters
        ----------
        expr: str
            An expression passed to the backend. An example could be a SQL or
            pandas query. Largely dependent on concrete implementations.
        **expr_args: dict (optional)
            Arguments to pass to the SQL expression. Should be in the form:
            {"arg_name": arg_value, ...}

        Returns
        ----------
        pandas.DataFrame of the resulting query.
        """
        expr_args = kwargs.get("expr_args", ())
        results, columns = self.execute_sql(expr, expr_args)
        results = [list(row) for row in results]
        return pd.DataFrame(results, columns=columns)

    def __len__(self):
        """Returns the number of baskets in the index."""
        results, _ = self.execute_sql(
                f"SELECT COUNT (*) FROM {self.pantry_schema}.pantry_index"
            )
        return results[0][0]

    def __str__(self):
        """Returns the str instantiation type of this Index (ie 'IndexSQL')."""
        return self.__class__.__name__



"""Wherein is contained functionality concerning validating the
basket_manifest.json file.
"""

import jsonschema
from jsonschema import validate

from ..config import manifest_schema


# Validate basket keys and value data types on read in
def validate_basket_dict(basket_dict):
    """Validate the basket_manifest.json has the correct structure.

    Parameters:
    ----------
    basket_dict: dict
        Dictionary of the basket contents
        read in from basket_manifest.json in the file system.

    Returns
    ----------
    bool: True if basket has correct schema, False otherwise.
    """

    try:
        validate(instance=basket_dict, schema=manifest_schema)
        return True

    except jsonschema.exceptions.ValidationError:
        return False



"""Pytest tests for the sqlite index."""
import os
import sqlite3

import pytest

from weave.pantry import Pantry
from weave.index.index_sqlite import IndexSQLite
from weave.tests.pytest_resources import get_sample_basket_df, get_file_systems
from weave.tests.pytest_resources import IndexForTest
from weave.tests.pytest_resources import PantryForTest


###############################################################################
#                      Pytest Fixtures Documentation:                         #
#            https://docs.pytest.org/en/7.3.x/how-to/fixtures.html            #
#                                                                             #
#                  https://docs.pytest.org/en/7.3.x/how-to/                   #
#          fixtures.html#teardown-cleanup-aka-fixture-finalization            #
#                                                                             #
#  https://docs.pytest.org/en/7.3.x/how-to/fixtures.html#fixture-parametrize  #
###############################################################################

# Pylint doesn't like redefining the test fixture here from
# test_basket, but this is the right way to do it if at some
# point in the future the two need to be differentiated.
# pylint: disable=duplicate-code

# Create fsspec objects to be tested, and add to file_systems list.
file_systems, file_systems_ids = get_file_systems()


# Test with different fsspec file systems (above).
@pytest.fixture(
    name="test_pantry",
    params=file_systems,
    ids=file_systems_ids,
)
def fixture_test_pantry(request, tmpdir):
    """Sets up test pantry for the tests"""
    file_system = request.param
    test_pantry = PantryForTest(tmpdir, file_system)
    yield test_pantry
    test_pantry.cleanup_pantry()


@pytest.fixture(
    name="test_index",
    params=[IndexSQLite],
    ids=["IndexSQLite"],
)
def fixture_test_index(request):
    """Sets up test index for the tests"""
    index_constructor = request.param
    test_index = IndexForTest(index_constructor, file_systems[0])
    yield test_index
    test_index.cleanup_index()


def test_index_two_pantries_with_same_name(test_pantry):
    """Validate that 2 pantry objects with the same basename have their
    own index.
    """
    pantry_1_path = os.path.join(test_pantry.pantry_path, "test-pantry-1")
    pantry_2_path = os.path.join(test_pantry.pantry_path,
                                 "test",
                                 "test-pantry-1")

    test_pantry.file_system.mkdir(pantry_1_path)
    test_pantry.file_system.mkdir(pantry_2_path)

    tmp_txt_file = test_pantry.tmpdir.join("test.txt")
    tmp_txt_file.write("this is a test")

    upload_path_1 = os.path.join(pantry_1_path, "text.txt")
    upload_path_2 = os.path.join(pantry_2_path, "text.txt")

    # Must upload a file because Minio will remove empty directories
    test_pantry.file_system.upload(str(tmp_txt_file.realpath()), upload_path_1)
    test_pantry.file_system.upload(str(tmp_txt_file.realpath()), upload_path_2)

    # Make the Pantries.
    pantry_1 = Pantry(
        IndexSQLite,
        pantry_path=pantry_1_path,
        file_system=test_pantry.file_system,
    )

    pantry_2 = Pantry(
        IndexSQLite,
        pantry_path=pantry_2_path,
        file_system=test_pantry.file_system,
    )

    # Because the two Pantry Objects have the same name, or path basename,
    # the db files must be different.
    assert pantry_1.index.db_path != pantry_2.index.db_path

    # Remove the .db files that are not cleaned up with 'test_pantry'
    pantry_1.index.cur.close()
    pantry_1.index.con.close()
    pantry_2.index.cur.close()
    pantry_2.index.con.close()
    os.remove(pantry_1.index.db_path)
    os.remove(pantry_2.index.db_path)


def test_index_uploaded_basket_not_found_in_another_index(test_pantry):
    """Validate that a basket uploaded to one pantry does not show up in
    another pantry.
    """
    pantry_1_path = os.path.join(test_pantry.pantry_path, "test-pantry-1")
    pantry_2_path = os.path.join(test_pantry.pantry_path, "test-pantry-2")

    test_pantry.file_system.mkdir(pantry_1_path)
    test_pantry.file_system.mkdir(pantry_2_path)

    tmp_txt_file = test_pantry.tmpdir.join("test.txt")
    tmp_txt_file.write("this is a test")

    upload_path_1 = os.path.join(pantry_1_path, "text.txt")
    upload_path_2 = os.path.join(pantry_2_path, "text.txt")

    # Must upload a file because Minio will remove empty directories
    test_pantry.file_system.upload(str(tmp_txt_file.realpath()), upload_path_1)
    test_pantry.file_system.upload(str(tmp_txt_file.realpath()), upload_path_2)

    # Make the Pantries.
    pantry_1 = Pantry(
        IndexSQLite,
        pantry_path=pantry_1_path,
        file_system=test_pantry.file_system,
    )

    pantry_2 = Pantry(
        IndexSQLite,
        pantry_path=pantry_2_path,
        file_system=test_pantry.file_system,
    )

    # Upload a basket to each pantry.
    pantry_1.upload_basket(
        upload_items=[{"path":str(tmp_txt_file.realpath()), "stub":False}],
        basket_type="test-basket",
        unique_id="0000",
    )
    pantry_2.upload_basket(
        upload_items=[{"path":str(tmp_txt_file.realpath()), "stub":False}],
        basket_type="test-basket",
        unique_id="0001",
    )

    # Check that the indices are not the same
    pantry_1_index = pantry_1.index.to_pandas_df()
    pantry_2_index = pantry_2.index.to_pandas_df()

    # Validate that each basket was uploaded to the proper pantry, and
    # not the other.
    assert not pantry_1_index.equals(pantry_2_index)
    assert "0000" in pantry_1_index["uuid"].values
    assert "0000" not in pantry_2_index["uuid"].values
    assert "0001" in pantry_2_index["uuid"].values
    assert "0001" not in pantry_1_index["uuid"].values

    # Remove the .db files that are not cleaned up with 'test_pantry'
    pantry_1.index.cur.close()
    pantry_1.index.con.close()
    pantry_2.index.cur.close()
    pantry_2.index.con.close()
    os.remove(pantry_1.index.db_path)
    os.remove(pantry_2.index.db_path)


def test_index_sqlite_track_basket_adds_to_parent_uuids(test_index):
    """Test that track_basket adds necessary rows to the parent_uuids table."""
    sample_basket_df = get_sample_basket_df()
    uuid = "1000"
    sample_basket_df["uuid"] = uuid

    # Add uuids to the parent_uuids of the df.
    sample_basket_df["parent_uuids"] = [["0001", "0002", "0003"]]

    # Track the basket.
    test_index.index.track_basket(sample_basket_df)

    # Use the cursor to check the parent_uuids table.
    cursor = test_index.index.con.cursor()
    cursor.execute("SELECT * FROM parent_uuids")
    rows = cursor.fetchall()

    # Check we have the expected values.
    assert len(rows) == 3
    assert rows[0] == (uuid, "0001")
    assert rows[1] == (uuid, "0002")
    assert rows[2] == (uuid, "0003")

    # Untrack the basket and ensure values are removed from parent_uuids table.
    test_index.index.untrack_basket(uuid)
    cursor.execute("SELECT * FROM parent_uuids")
    rows = cursor.fetchall()
    assert len(rows) == 0


def test_drop_index_deletes_sqlite_file(test_index):
    """Test that drop_index drops all data and deletes the db file."""
    sample_basket_df = get_sample_basket_df()
    uuid = "1000"
    sample_basket_df["uuid"] = uuid

    # Track the basket.
    test_index.index.track_basket(sample_basket_df)

    assert len(test_index.index) == 1

    # Drop the index (deleting the db file)
    test_index.index.drop_index()

    # Check the db is closed when we try to use it.
    with pytest.raises(sqlite3.ProgrammingError) as exc:
        len(test_index.index)
    assert "Cannot operate on a closed database." in str(exc)

    # Check the db file is deleted.
    assert not os.path.exists(test_index.db_path)

    # Recreate the dbfile using clear_index so the test doesn't crash during
    # cleanup, as we previously deleted the sqlite file.
    test_index.index.clear_index()



"""Pytest tests for the index directory."""
import os
import re
import tempfile
import warnings

import pytest

from weave.pantry import Pantry
from weave.index.index_pandas import IndexPandas
from weave.tests.pytest_resources import PantryForTest, get_file_systems


###############################################################################
#                      Pytest Fixtures Documentation:                         #
#            https://docs.pytest.org/en/7.3.x/how-to/fixtures.html            #
#                                                                             #
#                  https://docs.pytest.org/en/7.3.x/how-to/                   #
#          fixtures.html#teardown-cleanup-aka-fixture-finalization            #
#                                                                             #
#  https://docs.pytest.org/en/7.3.x/how-to/fixtures.html#fixture-parametrize  #
###############################################################################

# Pylint doesn't like redefining the test fixture here from
# test_basket, but this is the right way to do it if at some
# point in the future the two need to be differentiated.
# pylint: disable=duplicate-code

# Create fsspec objects to be tested, and add to file_systems list.
file_systems, file_systems_ids = get_file_systems()


# Test with different fsspec file systems (above).
@pytest.fixture(
    name="test_pantry",
    params=file_systems,
    ids=file_systems_ids,
)
def fixture_test_pantry(request, tmpdir):
    """Sets up test pantry for the tests"""
    file_system = request.param
    test_pantry = PantryForTest(tmpdir, file_system)
    yield test_pantry
    test_pantry.cleanup_pantry()


def test_sync_index_gets_latest_index(test_pantry):
    """Tests IndexPandas.sync_index by generating two distinct objects and
    making sure that they are both syncing to the index pandas DF (represented
    by JSON) on the file_system."""

    # Put basket in the temporary pantry
    tmp_basket_dir_one = test_pantry.set_up_basket("basket_one")
    test_pantry.upload_basket(tmp_basket_dir=tmp_basket_dir_one, uid="0001")

    # Create index
    pantry = Pantry(
        IndexPandas,
        pantry_path=test_pantry.pantry_path,
        file_system=test_pantry.file_system,
        sync=True,
    )
    pantry.index.to_pandas_df()

    # Add another basket
    tmp_basket_dir_two = test_pantry.set_up_basket("basket_two")
    test_pantry.upload_basket(tmp_basket_dir=tmp_basket_dir_two, uid="0002")

    # Regenerate index outside of current index object
    pantry2 = Pantry(
        IndexPandas,
        pantry_path=test_pantry.pantry_path,
        file_system=test_pantry.file_system,
        sync=True,
    )
    pantry2.index.generate_index()

    # assert length of index includes both baskets and excludes the index
    assert len(pantry.index.to_pandas_df()) == 2

    #assert all baskets in index are not index baskets
    for i in range(len(pantry.index.to_pandas_df())):
        basket_type = pantry.index.to_pandas_df()["basket_type"][i]
        assert basket_type != "index"


def test_sync_index_calls_generate_index_if_no_index(test_pantry):
    """Test to make sure that if there isn't a index available then
    generate_index will still be called."""

    # Put basket in the temporary pantry
    tmp_basket_dir_one = test_pantry.set_up_basket("basket_one")
    test_pantry.upload_basket(tmp_basket_dir=tmp_basket_dir_one, uid="0001")

    # Create index
    pantry = Pantry(
        IndexPandas,
        pantry_path=test_pantry.pantry_path,
        file_system=test_pantry.file_system,
        sync=True,
    )
    assert len(pantry.index.to_pandas_df()) == 1

def test_sync_index_autocleanup_if_false(test_pantry):
    """Tests that IndexPandas.sync raises the correct warning with
    auto-cleanup turned off."""
    pantry = Pantry(
        IndexPandas,
        pantry_path=test_pantry.pantry_path,
        file_system=test_pantry.file_system,
        sync=True,
        auto_cleanup=False,
    )

    for i in range (25):
        tmp_basket_dir_one = test_pantry.set_up_basket(f"basket_{i}")
        test_pantry.upload_basket(tmp_basket_dir=tmp_basket_dir_one,
                             uid=f"000{i}")

        pantry.index.generate_index()

    with warnings.catch_warnings(record=True) as warn:
        pantry.index.sync_index()
    warning_list = [warn[i].message for i in range(len(warn))]
    warning_1 = warning_list[0]
    assert warning_1.args[0] == (
        "The index basket count is 25. "
        "Consider running weave.IndexPandas.clean_up_indices"
    )

def test_get_index_time_from_path(test_pantry):
    """Tests Index._get_index_time_from_path to ensure it returns the correct
    string."""

    path = "C:/asdf/gsdjls/1234567890-index.json"
    # Testing a protected access var here.
    # pylint: disable-next=protected-access
    time = Pantry(
        IndexPandas,
        pantry_path=test_pantry.pantry_path,
        file_system=test_pantry.file_system,
        sync=True,
    ).index._get_index_time_from_path(path=path)
    assert time == 1234567890


def test_clean_up_indices_n_not_int(test_pantry):
    """Tests that IndexPandas.clean_up_indices errors on a str (should be int).
    """

    test_str = "the test"
    with pytest.raises(
        ValueError,
        match=re.escape("invalid literal for int() with base 10: 'the test'"),
    ):
        pantry = Pantry(
            IndexPandas,
            pantry_path=test_pantry.pantry_path,
            file_system=test_pantry.file_system,
            sync=True,
        )
        pantry.index.clean_up_indices(n_keep=test_str)


def test_clean_up_indices_leaves_n_indices(test_pantry):
    """Tests that IndexPandas.clean_up_indices leaves behind the correct number
    of indices."""

    # Put basket in the temporary pantry
    tmp_basket_dir_one = test_pantry.set_up_basket("basket_one")
    test_pantry.upload_basket(tmp_basket_dir=tmp_basket_dir_one, uid="0001")

    # Create index
    pantry = Pantry(
        IndexPandas,
        pantry_path=test_pantry.pantry_path,
        file_system=test_pantry.file_system,
        sync=True,
    )
    pantry.index.to_pandas_df()

    # Add another basket
    tmp_basket_dir_two = test_pantry.set_up_basket("basket_two")
    test_pantry.upload_basket(tmp_basket_dir=tmp_basket_dir_two, uid="0002")
    pantry.index.generate_index()

    # Now there should be two index baskets. clean up all but one of them:
    pantry.index.clean_up_indices(n_keep=1)
    index_path = os.path.join(test_pantry.pantry_path, "index")
    assert len(test_pantry.file_system.ls(index_path)) == 1


def test_clean_up_indices_with_n_greater_than_num_of_indices(test_pantry):
    """Tests that IndexPandas.clean_up_indices behaves well when given a number
    greater than the total number of indices."""

    # Put basket in the temporary pantry
    tmp_basket_dir_one = test_pantry.set_up_basket("basket_one")
    test_pantry.upload_basket(tmp_basket_dir=tmp_basket_dir_one, uid="0001")

    # Create index
    pantry = Pantry(
        IndexPandas,
        pantry_path=test_pantry.pantry_path,
        file_system=test_pantry.file_system,
        sync=True,
    )
    pantry.index.to_pandas_df()

    # Add another basket
    tmp_basket_dir_two = test_pantry.set_up_basket("basket_two")
    test_pantry.upload_basket(tmp_basket_dir=tmp_basket_dir_two, uid="0002")
    pantry.index.generate_index()

    # Now there should be two index baskets. clean up all but three of them:
    pantry.index.clean_up_indices(n_keep=3)
    index_path = os.path.join(test_pantry.pantry_path, "index")
    assert len(test_pantry.file_system.ls(index_path)) == 2


def test_is_index_current(test_pantry):
    """Creates two IndexPandas objects and pits them against eachother in order
    to ensure that IndexPandas.is_index_current is working as expected."""

    # Put basket in the temporary pantry
    tmp_basket_dir_one = test_pantry.set_up_basket("basket_one")
    test_pantry.upload_basket(tmp_basket_dir=tmp_basket_dir_one, uid="0001")

    # Create index
    pantry = Pantry(
        IndexPandas,
        pantry_path=test_pantry.pantry_path,
        file_system=test_pantry.file_system,
        sync=True,
    )
    pantry.index.to_pandas_df()

    # Add another basket
    tmp_basket_dir_two = test_pantry.set_up_basket("basket_two")
    test_pantry.upload_basket(tmp_basket_dir=tmp_basket_dir_two, uid="0002")

    # Regenerate index outside of current index object
    pantry2 = Pantry(
        IndexPandas,
        pantry_path=test_pantry.pantry_path,
        file_system=test_pantry.file_system,
        sync=True,
    )
    pantry2.index.generate_index()
    assert pantry2.index.is_index_current() is True
    assert pantry.index.is_index_current() is False

def test_index_updated_after_new_pantry_basket_upload(test_pantry):
    """Tests basket_upload updates the index after creating IndexPandas pantry
    object when another IndexPandas index exists."""

    # Create first pantry
    pantry = Pantry(
        IndexPandas,
        pantry_path=test_pantry.pantry_path,
        file_system=test_pantry.file_system,
        sync=True,
    )
    # Create temp directory for uploading basket
    with tempfile.TemporaryDirectory(dir=".") as tmpdir:
        tmp_file_path = os.path.join(tmpdir, "temp_basket.txt")
        with open(tmp_file_path, "w", encoding="utf-8") as tmp_file:
            pantry.upload_basket(
                upload_items=[{"path":tmp_file.name, "stub":False}],
                basket_type="test-1",
            )
    # Create second pantry
    pantry2 = Pantry(
        IndexPandas,
        pantry_path=test_pantry.pantry_path,
        file_system=test_pantry.file_system,
        sync=True,
    )
    # Upload basket with new index
    with tempfile.TemporaryDirectory(dir=".") as tmpdir:
        tmp_file_path = os.path.join(tmpdir, "temp_basket.txt")
        with open(tmp_file_path, "w", encoding="utf-8") as tmp_file:
            pantry2.upload_basket(
                upload_items=[{"path":tmp_file.name, "stub":False}],
                basket_type="test-1",
            )
    assert len(pantry2.index.to_pandas_df()) == 2



"""Contains scripts concerning Mongo Loader functionality."""

# Ignore pylint duplicate code. Code here is used to explicitly show pymongo is
# an optional dependency. Duplicate code is found in config.py (where pymongo
# is actually imported)
# pylint: disable-next=duplicate-code
try:
    import pymongo
except ImportError:
    _HAS_PYMONGO = False
else:
    _HAS_PYMONGO = True

from .basket import Basket
from .config import get_mongo_db

# pylint: disable-next=too-many-instance-attributes
class MongoLoader():
    """Initializes a mongo loader class. Retrieves a connection to the mongo
    server and facilitates the uploading of records to each pantry's individual
    mongo db based on the record type (ie supplement, manifest, metadata).
    """

    def __init__(self, pantry, mongo_client=None, **kwargs):
        """Creates the mongo loader and makes a reference to the pantry's DB.

        Parameters
        ----------
        pantry: weave.Pantry
            A pantry object
        mongo_client: pymongo.MongoClient or None (optional; defaults to None)
            A pre-constructed MongoClient to be used. Ignored if the Pantry arg
            has a non-None pantry.mongo_client. Otherwise, this arg will be
            used for the client. If none, is provided, a MongoClient will be
            constructed using the optional mongo_config dictionary,
            or weave.config.get_mongo_db() as a last resort.
        **mongo_config: dict (optional)
            Dictionary containing the configuration settings of this loader.
            Supported Keys:
        """
        if not _HAS_PYMONGO:
            raise ImportError("Missing Dependency. The package 'pymongo' "
                              "is required to use this class.")

        self.pantry = pantry
        self.mongo_config = kwargs.get("mongo_config", {})

        # Priorize using the pantry client, otherwise use the given client.
        self.mongo_client = self.pantry.mongo_client or mongo_client
        # If both were None, try to make a client using the mongo_config.
        if self.mongo_client is None and self.mongo_config:
            self.mongo_client = pymongo.MongoClient(
                host=self.mongo_config["mongodb_host"],
                username=self.mongo_config["mongodb_username"],
                password=self.mongo_config["mongodb_password"],
                port=self.mongo_config.get("mongodb_port", 27017),
                timeoutMS=self.mongo_config.get("mongodb_timeout", None)
            )
            self.mongo_client.server_info()
        # If we still don't have a valid mongo_client, use the weave config as
        # a last resort.
        if self.mongo_client is None:
            timeout = self.mongo_config.get("mongodb_timeout", None)
            self.mongo_client = get_mongo_db(timeout=timeout)

        # Get the database. (Use MONGODB_DATABASE, defaulting to
        # pantry_path if it is not present.)
        self.database_name = self.mongo_config.get(
            "mongodb_database", self.pantry.pantry_path)
        self.database = self.mongo_client[self.database_name]

        self.metadata_collection = self.mongo_config.get(
            "metadata_collection", "metadata")
        self.manifest_collection = self.mongo_config.get(
            "manifest_collection", "manifest")
        self.supplement_collection = self.mongo_config.get(
            "supplement_collection", "supplement")


    def load_mongo_metadata(self, uuids, **kwargs):
        """Load metadata from baskets into the mongo database.

        A metadata.json is created in baskets when the metadata
        field is provided upon upload. This metadata is added to the
        Mongo database when invoking load_mongo_metadata. UUID and basket_type
        from the index_table are also added to Mongo for referrence
        back to the datasource.

        Parameters
        ----------
        uuids: [str]
            A list of uuids to add their metadata to the mongo db.
        **collection: str (optional; defaults to self.metadata_collection)
            Metadata will be added to the Mongo collection specified. If not
            provided, populate using self.metadata_collection (which defaults
            to "metadata" if otherwise unspecified).
        """
        collection = kwargs.get("collection", self.metadata_collection)
        if not isinstance(uuids, list):
            uuids = [uuids]
        if not isinstance(uuids[0], str):
            raise TypeError("Invalid datatype for uuids: "
                            "must be a list of strings [str]")
        if not isinstance(collection, str):
            raise TypeError("Invalid datatype for metadata collection: "
                            "must be a string")

        for uuid in uuids:
            basket = Basket(uuid, pantry=self.pantry)
            metadata = basket.get_metadata()
            if not metadata:
                continue
            mongo_metadata = {}
            mongo_metadata["uuid"] = basket.uuid
            mongo_metadata["basket_type"] = basket.basket_type
            mongo_metadata.update(metadata)

            # If the UUID already has metadata loaded in MongoDB,
            # the metadata should not be loaded to MongoDB again.
            if 0 == self.database[
                self.metadata_collection
            ].count_documents({"uuid": basket.uuid}):
                self.database[collection].insert_one(mongo_metadata)

    def load_mongo_manifest(self, uuids, **kwargs):
        """Load manifest from baskets into the mongo database.

        A manifest.json is created in baskets upon upload. This manifest is
        added to the Mongo database when invoking load_mongo_manifest. UUID,
        and basket_type from the index_table are also added to Mongo for
        referrence back to the datasource.

        Parameters
        ----------
        uuids: [str]
            A list of uuids to add their manifests to the mongo db.
        **collection: str (optional; defaults to self.manifest_collection)
            Manifests will be added to the Mongo collection specified. If not
            provided, populate using self.manifest_collection (which defaults
            to "manifest" if otherwise unspecified).
        """
        collection = kwargs.get("collection", self.manifest_collection)
        if not isinstance(uuids, list):
            uuids = [uuids]
        if not isinstance(uuids[0], str):
            raise TypeError("Invalid datatype for uuids: "
                            "must be a list of strings [str]")
        if not isinstance(collection, str):
            raise TypeError("Invalid datatype for manifest collection: "
                            "must be a string")
        for uuid in uuids:
            basket = Basket(uuid, pantry=self.pantry)
            mongo_manifest = basket.get_manifest()
            if not mongo_manifest:
                continue

            # If the UUID already has the manifest loaded in MongoDB,
            # the manifest should not be loaded to MongoDB again.
            if 0 == self.database[
                collection
            ].count_documents({"uuid": basket.uuid}):
                self.database[collection].insert_one(mongo_manifest)

    def load_mongo_supplement(self, uuids, **kwargs):
        """Load supplement from baskets into the mongo database.

        A supplement.json is created in baskets upon upload. This supplement
        is added to the Mongo database when invoking load_mongo_supplement.
        UUID, and basket_type from the index_table are also added to Mongo for
        referrence back to the datasource.

        Parameters
        ----------
        uuids: [str]
            A list of uuids to add their supplement to the mongo db.
        **collection: str (optional; defaults to self.supplement_collection)
            Supplements will be added to the Mongo collection specified. If not
            provided, populate using self.supplement_collection (which defaults
            to "supplement" if otherwise unspecified).
        """
        collection = kwargs.get("collection", self.supplement_collection)
        if not isinstance(uuids, list):
            uuids = [uuids]
        if not isinstance(uuids[0], str):
            raise TypeError("Invalid datatype for uuids: "
                            "must be a list of strings [str]")
        if not isinstance(collection, str):
            raise TypeError("Invalid datatype for supplement collection: "
                            "must be a string")

        for uuid in uuids:
            basket = Basket(uuid, pantry=self.pantry)
            supplement = basket.get_supplement()
            if not supplement:
                continue
            mongo_supplement = {}
            mongo_supplement["uuid"] = basket.uuid
            mongo_supplement["basket_type"] = basket.basket_type
            mongo_supplement.update(supplement)

            # If the UUID already has metadata loaded in MongoDB,
            # the metadata should not be loaded to MongoDB again.
            if 0 == self.database[
                collection
            ].count_documents({"uuid": basket.uuid}):
                self.database[collection].insert_one(mongo_supplement)


    def load_mongo(self, uuids, **kwargs):
        """Load metadata, manifest, and supplement from baskets into the
        mongo database.

        Parameters
        ----------
        uuids: [str]
            A list of uuids to add their data to the mongo db.
        **metadata_collection: str (default=self.metadata_collection)
            Metadata will be added to the Mongo collection specified.
        **manifest_collection: str (default=self.manifest_collection)
            Manifest will be added to the Mongo collection specified.
        **supplement_collection: str (default=self.supplement_collection)
            Supplement will be added to the Mongo collection specified.
        """
        metadata_collection = kwargs.get("metadata_collection",
                                         self.metadata_collection)
        manifest_collection = kwargs.get("manifest_collection",
                                         self.manifest_collection)
        supplement_collection = kwargs.get("supplement_collection",
                                           self.supplement_collection)

        self.load_mongo_metadata(uuids, collection=metadata_collection)
        self.load_mongo_manifest(uuids, collection=manifest_collection)
        self.load_mongo_supplement(uuids, collection=supplement_collection)

    def clear_mongo(self, refresh=False):
        """Clear the metadata, manifest, and supplement collections optionally,
        refreshing them from the pantry.

        Parameters
        ----------
        refresh: bool (default=False)
            If True, reload the collections with data retreived from the pantry
        """
        # Drop the referenced database.
        self.database.client.drop_database(self.database_name)

        # Optionally refresh the mongo collections with the current index.
        if refresh:
            self.load_mongo(
                self.pantry.index.to_pandas_df(max_rows=None)['uuid'].to_list()
            )

    def remove_document(self, uuid, **kwargs):
        """Delete all documents containing the given uuid from all collections.

        Parameters
        ----------
        uuid: str
            "uuid" will be used to locate and remove the documents from MongoDB
            in the supplement, manifest, and metadata collections.
        **metadata_collection: str (default=self.metadata_collection)
            Metadata will be added to the Mongo collection specified.
        **manifest_collection: str (default=self.manifest_collection)
            Manifest will be added to the Mongo collection specified.
        **supplement_collection: str (default=self.supplement_collection)
            Supplement will be added to the Mongo collection specified.
        """
        metadata_collection = kwargs.get("metadata_collection",
                                         self.metadata_collection)
        manifest_collection = kwargs.get("manifest_collection",
                                         self.manifest_collection)
        supplement_collection = kwargs.get("supplement_collection",
                                           self.supplement_collection)

        collection_names = (metadata_collection,
                            manifest_collection,
                            supplement_collection)

        for collection in collection_names:
            if not isinstance(collection, str):
                raise TypeError("Invalid datatype for collection: "
                                "must be a string")

        for collection in collection_names:
            self.database[collection].delete_many({'uuid':uuid})



"""Wherein is contained the Pantry class.
"""
import json
import os
import warnings

import s3fs
# Ignore pylint duplicate code. Code here is used to explicitly show pymongo is
# an optional dependency. Duplicate code is found in config.py (where pymongo
# is actually imported)
# pylint: disable-next=duplicate-code
try:
    # For the sake of explicitly showing that pymongo is optional, import
    # pymongo here, even though it is not currently used in this file.
    # Pylint ignore the next unused-import pylint warning.
    # Also inline ruff ignore unused import (F401)
    # pylint: disable-next=unused-import
    import pymongo  # noqa: F401
except ImportError:
    _HAS_PYMONGO = False
else:
    _HAS_PYMONGO = True

from .mongo_loader import MongoLoader
from .basket import Basket
from .config import get_file_system
from .index.create_index import create_index_from_fs
from .index.index_abc import IndexABC
from .upload import UploadBasket, derive_integrity_data
from .validate import validate_pantry

# pylint: disable-next=too-many-instance-attributes
class Pantry():
    """Facilitate user interaction with the index of a Weave data warehouse.
    """

    def __init__(self, index: IndexABC, pantry_path="weave-test",
                 mongo_client=None, **kwargs):
        """Initialize Pantry object.

        A pantry is a collection of baskets. This class facilitates the upload,
        retrieval, and deletion of baskets within a file system. It uses and
        updates an index to track the contents within the pantry for rapid
        operations.

        Parameters:
        -----------
        index: IndexABC
            The concrete implementation of an IndexABC. This is used to track
            the contents within the pantry.
        pantry_path: str (default="weave-test")
            Name of the pantry this object is associated with.
        mongo_client: pymongo.MongoClient (optional)
            The MongoClient object to be used to update a mongo datastore when
            adding and removing baskets.
        **file_system: fsspec object (optional)
            The fsspec object which hosts the pantry we desire to index.
            If file_system is None, then the default fs is retrieved from the
            config.
        **mongo_config: dict (optional)
            A dictionary containing keys and values to be used for the mongo
            collection. Supported keys: MONGODB_HOST, MONGODB_PORT,
            MONGODB_USERNAME, MONGODB_PASSWORD, METADATA_COLLECTION,
            MANIFEST_COLLECTION, SUPPLEMENT_COLLECTION.
            If keys are not present, default values will be used if necessary.
        """
        self.file_system = kwargs.pop("file_system", None)
        if self.file_system is None:
            self.file_system = get_file_system()

        # If using S3FileSystem, check the connection
        if isinstance(self.file_system, s3fs.S3FileSystem):
            try:
                self.file_system.ls('s3://')
            except Exception as exc:
                raise ConnectionError("Connection to s3fs failed.") from exc

        try:
            if not self.file_system.exists(pantry_path):
                self.file_system.mkdir(pantry_path)
        except Exception as exc:
            raise OSError("Failed to create directory, Invalid"
                            f" Path {pantry_path}.\n{exc}") from exc

        self.pantry_path = str(pantry_path)
        self.setup_config = {}
        self.load_setup_config()

        # Check if file system is read-only. If so, raise error.
        try:
            self.file_system.touch(os.path.join(self.pantry_path,
                                                "test_read_only.txt"))
            self.file_system.rm(os.path.join(self.pantry_path,
                                                "test_read_only.txt"))
            self.is_read_only = False
        except (OSError, ValueError):
            self.is_read_only = True

        self.index = index(
            file_system=self.file_system,
            pantry_path=self.pantry_path,
            setup_config=self.setup_config['index_setup_config'],
            pantry_read_only=self.is_read_only,
            **kwargs,
        )
        self.mongo_client = mongo_client
        self.mongo_config = kwargs.get("mongo_config", {})
        self.setup_config['index_setup_config'] = self.index.generate_config()
        self._generate_config()

    def validate_path_in_pantry(self, path):
        """Validate the given path is within the pantry.

        Check 1: Ensure the path begins with the pantry path.
        Check 2: Ensure the ".." navigation command is not used.

        Parameters:
        -----------
        path: str
            Path to verify.
        """
        valid = True
        if self.pantry_path:
            valid = path.startswith(self.pantry_path + os.path.sep)
        if valid and "zip" not in str(type(self.file_system)):
            bad_str = os.path.sep + ".." + os.path.sep
            valid = bad_str not in path

        if not valid:
            raise ValueError(
                f"Attempting to access basket outside of pantry: {path}"
            )

    def load_metadata(self):
        """(Deprecated) Load pantry metadata from config.json."""
        warnings.warn(
            UserWarning("This function is being deprecated in a future weave "
                        "release. Please use load_setup_config() instead.")
        )
        self.load_setup_config()

    def load_setup_config(self):
        """Load pantry setup configuration from config.json."""
        self.config_path = os.path.join(
            self.pantry_path,
            'config.json',
        )
        if self.file_system.exists(self.config_path):
            with self.file_system.open(self.config_path, "rb") as file:
                self.setup_config = json.load(file)
        else:
            self.setup_config = {}
        if 'index_setup_config' not in self.setup_config:
            self.setup_config['index_setup_config'] = {}

    def _generate_config(self):
        """Helper method to populate the setup_config dictionary."""
        self.setup_config["index"] = str(self.index)
        self.setup_config["file_system"] = str(
            self.file_system.__class__.__name__
        )
        self.setup_config["pantry_path"] = self.pantry_path

        # Add the provided mongo_config to the main config dictionary.
        if self.mongo_config:
            self.setup_config.update(self.mongo_config)

    def save_metadata(self):
        """(Deprecated) Dump metadata to to pantry metadata file."""
        warnings.warn(
            UserWarning("This function is being deprecated in a future weave "
                        "release. Please use save_setup_config() instead.")
        )
        self.save_setup_config()

    def save_setup_config(self):
        """Save setup configuration settings to a config file inside the
        pantry."""
        with self.file_system.open(
            self.config_path, "w", encoding="utf-8"
        ) as outfile:
            json.dump(self.setup_config, outfile)

    def validate(self):
        """Convenient wrapper function to validate the pantry.

        Returns
        ----------
        A list of all invalid basket locations (will return an empty list if
        no warnings are raised)
        """

        return validate_pantry(self)

    def delete_basket(self, basket_address, **kwargs):
        """Deletes basket of given UUID or path.

        Note that the given basket will not be deleted if the basket is listed
        as the parent uuid for any of the baskets in the index.

        Parameters:
        -----------
        basket_address: str
            Argument can take one of two forms: either a path to the basket
            directory, or the UUID of the basket.
        **kwargs:
            Additional parameters to pass to the index
        """

        basket_address = str(basket_address)
        remove_item = self.index.get_rows(basket_address)

        if len(self.index.get_children(remove_item.iloc[0].uuid)) > 0:
            raise ValueError(
                f"The provided value for basket_uuid {basket_address} " +
                "is listed as a parent UUID for another basket. Please " +
                "delete that basket before deleting its parent basket."
            )

        self.validate_path_in_pantry(remove_item.iloc[0].address)
        self.index.untrack_basket(remove_item.iloc[0].address, **kwargs)
        self.file_system.rm(remove_item.iloc[0].address, recursive=True)

        if self.mongo_client is not None:
            MongoLoader(self).remove_document(remove_item.iloc[0].uuid)

    def upload_basket(self, upload_items, basket_type, **kwargs):
        """Upload a basket to the same pantry referenced by the Index

        Parameters
        ----------
        upload_items : [dict]
            List of python dictionaries with the following schema:
            {
                'path': path to the file or folder being uploaded (string),
                'stub': true/false (bool)
            }
            'path' can be a file or folder to be uploaded. Every filename
            and folder name must be unique. If 'stub' is set to True, integrity
            data will be included without uploading the actual file or folder.
            Stubs are useful when original file source information is desired
            without uploading the data itself. This is especially useful when
            dealing with large files.
        basket_type: str
            Type of basket being uploaded.
        **parent_ids: [str] (optional)
            List of unique ids associated with the parent baskets
            used to derive the new basket being uploaded.
        **metadata: dict (optional)
            Python dictionary that will be written to metadata.json
            and stored in the basket in upload file_system.
        **label: str (optional)
            Optional user friendly label associated with the basket.
        """
        # Check if file system is read-only. If so, raise error.
        if self.is_read_only:
            raise ValueError(
                "Unable to upload a basket to a read-only file system."
            )

        parent_ids = kwargs.pop("parent_ids", [])
        metadata = kwargs.pop("metadata", {})
        label = kwargs.pop("label", "")

        up_dir = UploadBasket(
            upload_items=upload_items,
            basket_type=basket_type,
            file_system=self.file_system,
            pantry_path=self.pantry_path,
            parent_ids=parent_ids,
            metadata=metadata,
            label=label,
            **kwargs,
        ).get_upload_path()

        single_indice_index = create_index_from_fs(up_dir, self.file_system)
        self.index.track_basket(single_indice_index)

        if self.mongo_client is not None:
            MongoLoader(self).\
                load_mongo(single_indice_index.iloc[0].uuid)

        return single_indice_index

    def get_basket(self, basket_address):
        """Retrieves a basket of given UUID or path.

        Parameters
        ----------
        basket_address: str
            Argument can take one of two forms: either a path to the Basket
            directory, or the UUID of the basket.

        Returns
        ----------
        The Basket object associated with the given UUID or path.
        """

        # Create a Basket from the given address, and the index's file_system
        # and pantry name. Basket will catch invalid inputs and raise
        # appropriate errors.
        row = self.index.get_rows(basket_address)
        if len(row) == 0:
            raise ValueError(f"Basket does not exist: {basket_address}")
        self.validate_path_in_pantry(row.iloc[0].address)
        return Basket(row.iloc[0].address, pantry=self)

    def does_file_exist(self, file_path, **kwargs):
        """Check if a file already exists inside the pantry and return
        the uuids where it does.

        Parameters
        ----------
        file_path: str
            Local path to file being checked
        **FORCE_LOAD_SUPPLEMENT: bool (default=false)
            If true, the mongo supplement will be reloaded with the ENTIRETY of
            the current pantry contents.

        Returns
        ----------
        list of basket uuids of where the file exists if it does. If file does
        not exist in the pantry, an empty list is returned.
        """
        if not _HAS_PYMONGO:
            raise ImportError("Missing Dependency. The package 'pymongo' "
                              "is required to use this function.")

        integrity_data = derive_integrity_data(file_path)['hash']
        try:
            mongo_loader = MongoLoader(pantry=self)
        except pymongo.errors.ServerSelectionTimeoutError:
            return []

        if kwargs.get("FORCE_LOAD_SUPPLEMENT", False):
            mongo_loader.load_mongo_supplement(
                self.index.to_pandas_df().uuid.to_list()
            )

        supplement_collection = mongo_loader.database['supplement']

        uuids = [f['uuid'] for f in supplement_collection.find(
            {'integrity_data.hash': {'$in': [integrity_data]}},
            {'uuid':1, '_id':0}
        )]

        return uuids



"""Pytests for the metadata_db functionality."""

import os
import tempfile
from unittest import mock

import pytest
from fsspec.implementations.local import LocalFileSystem

import weave
from weave import Pantry, IndexPandas
from weave.mongo_loader import MongoLoader
from weave.tests.pytest_resources import (PantryForTest, get_file_systems,
    get_pymongo_skip_reason, get_pymongo_skip_condition)

class MongoForTest(PantryForTest):
    """Extend the PantryForTest class to support mongodb and custom data
    loader.
    """
    def __init__(self, tmpdir, file_system):
        super().__init__(tmpdir, file_system)
        self.database = weave.config.get_mongo_db()[self.pantry_path]
        self.supplement_collection = "supplement"
        self.manifest_collection = "manifest"
        self.metadata_collection = "metadata"
        self.load_data()
        self.pantry = weave.Pantry(
            IndexPandas,
            pantry_path=self.pantry_path,
            file_system=self.file_system
        )
        self.pantry.index.generate_index()

    def load_data(self):
        """Loads data into the file_system."""
        # Create a temporary basket with a test file.
        tmp_basket_dir_name = "test_basket_tmp_dir"
        tmp_basket_dir = self.set_up_basket(tmp_basket_dir_name)

        # Upload the basket with different uuids and metadata.
        self.upload_basket(
            tmp_basket_dir, uid="1234", metadata={"key1": "value1"}
        )

        tmp_nested_dir = self.add_lower_dir_to_temp_basket(tmp_basket_dir)
        self.upload_basket(
            tmp_nested_dir, uid="4321", metadata={"key2": "value2"}
        )

        self.upload_basket(tmp_basket_dir, uid="nometadata")

    def cleanup(self):
        """Cleans up the pantry and mongodb."""
        self.cleanup_pantry()
        for collection in self.database.list_collection_names():
            self.database[collection].drop()
        self.database.client.drop_database(self.pantry_path)


# Create fsspec objects to be tested, and add to file_systems list.
file_systems, file_systems_ids = get_file_systems()

# Test with different fsspec file systems (above).
@pytest.fixture(
    params=file_systems,
    ids=file_systems_ids,
)
def set_up(request, tmpdir):
    """Sets up the fixture for testing usage."""
    file_system = request.param
    database = MongoForTest(tmpdir, file_system)
    yield database
    database.cleanup()


# Ignoring pylint's warning "redefined-outer-name" as this is simply
# how pytest works when it comes to pytest fixtures.
# pylint: disable=redefined-outer-name
@pytest.mark.skipif(
    get_pymongo_skip_condition(), reason=get_pymongo_skip_reason()
)
def test_load_mongo_metadata(set_up):
    """Test that load_mongo_metadata successfully loads valid metadata to
    the set_up.
    """
    uuids = ["1234", "4321", "nometadata"]
    mongo_loader = MongoLoader(pantry=set_up.pantry)
    mongo_loader.load_mongo_metadata(
        uuids=uuids,
        collection=set_up.metadata_collection,
    )

    truth_db = [
        {"uuid": "1234", "basket_type": "test_basket", "key1": "value1"},
        {"uuid": "4321", "basket_type": "test_basket", "key2": "value2"},
    ]

    db_data = list(set_up.database[set_up.metadata_collection].find({}))
    compared_data = []
    for item in db_data:
        item.pop("_id")
        compared_data.append(item)
    assert truth_db == compared_data


@pytest.mark.skipif(
    get_pymongo_skip_condition(), reason=get_pymongo_skip_reason()
)
def test_load_mongo_manifest(set_up):
    """Test that load_mongo_manifest successfully loads valid manifest to
    the set_up.
    """
    uuids = ["1234", "4321", "nometadata"]
    mongo_loader = MongoLoader(pantry=set_up.pantry)
    mongo_loader.load_mongo_manifest(
        uuids=uuids,
        collection=set_up.manifest_collection,
    )

    truth_db = [
        {"uuid": "1234", "parent_uuids": [], "basket_type": "test_basket",
         "label": ""},
        {"uuid": "4321", "parent_uuids": [], "basket_type": "test_basket",
         "label": ""},
        {"uuid": "nometadata", "parent_uuids": [],
         "basket_type": "test_basket", "label": ""}
    ]
    db_data = list(set_up.database[set_up.manifest_collection].find({}))
    compared_data = []
    for item in db_data:
        item.pop("_id")
        item.pop("upload_time")
        item.pop("weave_version")
        compared_data.append(item)
    assert truth_db == compared_data


@pytest.mark.skipif(
    get_pymongo_skip_condition(), reason=get_pymongo_skip_reason()
)
def test_load_mongo_supplement(set_up):
    """Test that load_mongo_supplement successfully loads valid supplement to
    the set_up.
    """
    uuids = ["1234", "4321", "nometadata"]
    mongo_loader = MongoLoader(pantry=set_up.pantry)
    mongo_loader.load_mongo_supplement(
        uuids=uuids,
        collection=set_up.supplement_collection,
    )

    truth_db = [
        {"uuid": "1234", "basket_type": "test_basket"},
        {"uuid": "4321", "basket_type": "test_basket"},
        {"uuid": "nometadata", "basket_type": "test_basket"}
    ]
    db_data = list(set_up.database[set_up.supplement_collection].find({}))
    compared_data = []
    for item in db_data:
        item.pop("_id")
        item.pop("integrity_data")
        item.pop("upload_items")
        compared_data.append(item)
    assert truth_db == compared_data


@pytest.mark.skipif(
    get_pymongo_skip_condition(), reason=get_pymongo_skip_reason()
)
def test_load_mongo(set_up):
    """Test that load_mongo successfully loads valid metadata, manifest, and
    supplement to the set_up.
    """
    uuids = ["1234", "4321", "nometadata"]
    mongo_loader = MongoLoader(pantry=set_up.pantry)
    mongo_loader.load_mongo(
        uuids=uuids,
        metadata_collection=set_up.metadata_collection,
        manifest_collection=set_up.manifest_collection,
        supplement_collection=set_up.supplement_collection,
    )

    metadata_truth_db = [
        {"uuid": "1234", "basket_type": "test_basket", "key1": "value1"},
        {"uuid": "4321", "basket_type": "test_basket", "key2": "value2"},
    ]
    metadata = list(set_up.database[set_up.metadata_collection].find({}))
    compared_metadata = []
    for item in metadata:
        item.pop("_id")
        compared_metadata.append(item)

    manifest_truth_db = [
        {"uuid": "1234", "parent_uuids": [], "basket_type": "test_basket",
         "label": ""},
        {"uuid": "4321", "parent_uuids": [], "basket_type": "test_basket",
         "label": ""},
        {"uuid": "nometadata", "parent_uuids": [],
         "basket_type": "test_basket", "label": ""}
    ]
    manifest = list(set_up.database[set_up.manifest_collection].find({}))
    compared_manifest = []
    for item in manifest:
        item.pop("_id")
        item.pop("upload_time")
        item.pop("weave_version")
        compared_manifest.append(item)

    supplement_truth_db = [
        {"uuid": "1234", "basket_type": "test_basket"},
        {"uuid": "4321", "basket_type": "test_basket"},
        {"uuid": "nometadata", "basket_type": "test_basket"}
    ]
    supplement = list(set_up.database[set_up.supplement_collection].find({}))
    compared_supplement = []
    for item in supplement:
        item.pop("_id")
        item.pop("integrity_data")
        item.pop("upload_items")
        compared_supplement.append(item)

    assert metadata_truth_db == compared_metadata
    assert manifest_truth_db == compared_manifest
    assert supplement_truth_db == compared_supplement


@pytest.mark.skipif(
    get_pymongo_skip_condition(), reason=get_pymongo_skip_reason()
)
def test_load_mongo_metadata_check_collection_for_string(set_up):
    """Test that load_mongo_metadata prevents loading data with an invalid
    set_up collection.
    """
    with pytest.raises(
        TypeError, match="Invalid datatype for metadata collection: "
                         "must be a string"
    ):
        mongo_loader = MongoLoader(pantry=set_up.pantry)
        mongo_loader.load_mongo_metadata(uuids=["1234"], collection=1)


@pytest.mark.skipif(
    get_pymongo_skip_condition(), reason=get_pymongo_skip_reason()
)
def test_load_mongo_metadata_check_for_duplicate_uuid(set_up):
    """Test duplicate metadata won't be uploaded to mongoDB, based on the UUID.
    """
    test_uuid = "1234"

    mongo_loader = MongoLoader(pantry=set_up.pantry)
    mongo_loader.load_mongo_metadata(
        [test_uuid],
        collection=set_up.metadata_collection,
    )
    mongo_loader.load_mongo_metadata(
        [test_uuid],
        collection=set_up.metadata_collection,
    )

    count = set_up.database[set_up.metadata_collection].count_documents(
        {"uuid": test_uuid}
    )
    assert count == 1, "duplicate uuid inserted"


@pytest.mark.skipif(
    get_pymongo_skip_condition(), reason=get_pymongo_skip_reason()
)
def test_load_mongo_manifest_check_for_duplicate_uuid(set_up):
    """Test duplicate manifest won't be uploaded to mongoDB, based on the UUID.
    """
    test_uuid = "1234"

    mongo_loader = MongoLoader(pantry=set_up.pantry)
    mongo_loader.load_mongo_manifest(
        [test_uuid],
        collection=set_up.manifest_collection,
    )
    mongo_loader.load_mongo_manifest(
        [test_uuid],
        collection=set_up.manifest_collection,
    )

    count = set_up.database[set_up.manifest_collection].count_documents(
        {"uuid": test_uuid}
    )
    assert count == 1, "duplicate uuid inserted"


@pytest.mark.skipif(
    get_pymongo_skip_condition(), reason=get_pymongo_skip_reason()
)
def test_load_mongo_supplement_check_for_duplicate_uuid(set_up):
    """Test duplicate supplement won't be uploaded to mongoDB, based on the
    UUID.
    """
    test_uuid = "1234"

    mongo_loader = MongoLoader(pantry=set_up.pantry)
    mongo_loader.load_mongo_supplement(
        [test_uuid],
        collection=set_up.supplement_collection,
    )
    mongo_loader.load_mongo_supplement(
        [test_uuid],
        collection=set_up.supplement_collection,
    )

    count = set_up.database[set_up.supplement_collection].count_documents(
        {"uuid": test_uuid}
    )
    assert count == 1, "duplicate uuid inserted"


@pytest.mark.skipif(
    get_pymongo_skip_condition(), reason=get_pymongo_skip_reason()
)
def test_check_file_already_exists(set_up):
    """Make a file, upload it to the pantry, check if that file already exists.
    """
    pantry = set_up.pantry

    with tempfile.NamedTemporaryFile() as tmp_file:
        tmp_file.write(b'This is my temporary file that we will hash')
        tmp_file.flush()
        pantry.upload_basket(
            upload_items=[{'path':tmp_file.name, 'stub':False}],
            basket_type='filealreadyexists',
            unique_id='file_already_exists_uuid',
        )

        mongo_loader = MongoLoader(pantry=pantry)
        mongo_loader.load_mongo_supplement(uuids=["file_already_exists_uuid"])
        uuids = pantry.does_file_exist(tmp_file.name)

        assert len(uuids) == 1
        assert uuids[0] == 'file_already_exists_uuid'


@pytest.mark.skipif(
    get_pymongo_skip_condition(), reason=get_pymongo_skip_reason()
)
@mock.patch.dict(os.environ, os.environ.copy(), clear=True)
def test_check_file_exists_no_mongodb(set_up):
    """Make a file, upload it to the pantry, check if that file already exists.
    """
    pantry = set_up.pantry

    with tempfile.NamedTemporaryFile() as tmp_file:
        tmp_file.write(b'This is my temporary file that we will hash')
        tmp_file.flush()
        pantry.upload_basket(
            upload_items=[{'path':tmp_file.name, 'stub':False}],
            basket_type='filealreadyexists',
            unique_id='file_already_exists_uuid',
        )

        mongo_loader = MongoLoader(pantry=pantry)
        mongo_loader.load_mongo_supplement(
            uuids=["file_already_exists_uuid"]
        )
        os.environ['MONGODB_HOST'] = "BAD_HOST"
        uuids = pantry.does_file_exist(tmp_file.name)
        assert len(uuids) == 0


@pytest.mark.skipif(
    get_pymongo_skip_condition(), reason=get_pymongo_skip_reason()
)
def test_check_pantries_have_discrete_mongodbs():
    """Create two pantries and check the databases are different
    when using does_file_exist.
    """
    localfs = LocalFileSystem()
    pantry_basename = (
        "pytest-temp-pantry"
        f"{os.environ.get('WEAVE_PYTEST_SUFFIX', '')}"
    )
    pantry_path1 = pantry_basename+"pantry1"
    pantry_path2 = pantry_basename+"pantry2"
    localfs.mkdir(pantry_path1)
    localfs.mkdir(pantry_path2)
    pantry1 = Pantry(
        IndexPandas,
        pantry_path=pantry_path1,
        file_system=localfs,
    )
    pantry2 = Pantry(
        IndexPandas,
        pantry_path=pantry_path2,
        file_system=localfs,
    )

    p1_mongo_loader = MongoLoader(pantry=pantry1)
    p2_mongo_loader = MongoLoader(pantry=pantry2)

    with tempfile.NamedTemporaryFile() as tmp_file1:
        tmp_file1.write(b'This is temporary file that we will hash for p1')
        tmp_file1.flush()
        with tempfile.NamedTemporaryFile() as tmp_file2:
            tmp_file2.write(b'This is temporary file that we will hash for p2')
            tmp_file2.flush()

            pantry1.upload_basket(
                upload_items=[{'path': tmp_file1.name, 'stub': False}],
                basket_type='test_basket',
                unique_id='pantry1fileuuid',
            )
            p1_mongo_loader.load_mongo_supplement(uuids=["pantry1fileuuid"])

            pantry2.upload_basket(
                upload_items=[{'path': tmp_file2.name, 'stub': False}],
                basket_type='test_basket',
                unique_id='pantry2fileuuid',
            )
            p2_mongo_loader.load_mongo_supplement(uuids=["pantry2fileuuid"])

            # Check the files only exist and are tracked according to their
            # respective pantries.
            assert (
                pantry1.does_file_exist(tmp_file1.name) == ["pantry1fileuuid"]
            )
            assert pantry1.does_file_exist(tmp_file2.name) == []
            assert pantry2.does_file_exist(tmp_file1.name) == []
            assert (
                pantry2.does_file_exist(tmp_file2.name) == ["pantry2fileuuid"]
            )

    # Manually clean up as we had to create our own pantries and mongodbs.
    p1_mongo_loader.database.client.drop_database(pantry1.pantry_path)
    p2_mongo_loader.database.client.drop_database(pantry2.pantry_path)
    localfs.rm(pantry1.pantry_path, recursive=True)
    localfs.rm(pantry2.pantry_path, recursive=True)


@pytest.mark.skipif(
    get_pymongo_skip_condition(), reason=get_pymongo_skip_reason()
)
def test_clear_mongo_works(set_up):
    """Test clear_mongo removed all mongo documents from the tracked DB and
    refreshed.
    """
    # Load the metadata, supplement, and manifests of the three baskets in
    # the pantry (one has no metadata).
    test_uuids = ["1234", "4321", "nometadata"]
    mongo_loader = MongoLoader(pantry=set_up.pantry)
    mongo_loader.load_mongo(uuids=test_uuids)

    count = set_up.database[set_up.metadata_collection].count_documents({})
    assert count == 2, "metadata collection not loaded"
    count = set_up.database[set_up.supplement_collection].count_documents({})
    assert count == 3, "supplement collection not loaded"
    count = set_up.database[set_up.manifest_collection].count_documents({})
    assert count == 3, "manifest collection not loaded"

    # Clear the database, and ensure all collections are empty.
    mongo_loader.clear_mongo()

    count = set_up.database[set_up.metadata_collection].count_documents({})
    assert count == 0, "metadata collection not cleared"
    count = set_up.database[set_up.supplement_collection].count_documents({})
    assert count == 0, "supplement collection not cleared"
    count = set_up.database[set_up.manifest_collection].count_documents({})
    assert count == 0, "manifest collection not cleared"

    # Refresh the database (metadata will only have 2 records as one of the
    # baskets does not include metadata)
    mongo_loader.clear_mongo(refresh=True)

    count = set_up.database[set_up.metadata_collection].count_documents({})
    assert count == 2, "metadata collection not up to date"
    count = set_up.database[set_up.supplement_collection].count_documents({})
    assert count == 3, "supplement collection not up to date"
    count = set_up.database[set_up.manifest_collection].count_documents({})
    assert count == 3, "manifest collection not up to date"


@pytest.mark.skipif(
    get_pymongo_skip_condition(), reason=get_pymongo_skip_reason()
)
def test_mongo_loader_remove_document_works(set_up):
    """Test remove document removes all documents containing the given uuid.
    """
    test_uuid = "1234"

    dummy_data = {
        "uuid": test_uuid,
        "dummy_data": "fake_data"
    }

    mongo_loader = MongoLoader(pantry=set_up.pantry)
    mongo_loader.load_mongo(
        uuids=[test_uuid],
        metadata_collection=set_up.metadata_collection,
        manifest_collection=set_up.manifest_collection,
        supplement_collection=set_up.supplement_collection,
    )

    metadata_collection = set_up.database[set_up.metadata_collection]
    manifest_collection = set_up.database[set_up.manifest_collection]
    supplement_collection = set_up.database[set_up.supplement_collection]

    # One record should exist in each collection after calling load_mongo.
    assert 1 == metadata_collection.count_documents({"uuid": test_uuid})
    assert 1 == manifest_collection.count_documents({"uuid": test_uuid})
    assert 1 == supplement_collection.count_documents({"uuid": test_uuid})

    # Manually insert a dummy record into each collection using the test uuid.
    metadata_collection.insert_one(dummy_data)
    supplement_collection.insert_one(dummy_data)
    manifest_collection.insert_one(dummy_data)

    # Check that the record was added successfully to each collection.
    assert 2 == metadata_collection.count_documents({"uuid": test_uuid})
    assert 2 == manifest_collection.count_documents({"uuid": test_uuid})
    assert 2 == supplement_collection.count_documents({"uuid": test_uuid})

    # Remove the test uuid from all the collections.
    mongo_loader.remove_document(uuid=test_uuid)

    # Every record (with that uuid) should be deleted from each collection.
    assert 0 == metadata_collection.count_documents({"uuid": test_uuid})
    assert 0 == manifest_collection.count_documents({"uuid": test_uuid})
    assert 0 == supplement_collection.count_documents({"uuid": test_uuid})



"""A Pantry Factory to build a Pantry object from default parameters or from a
pre-existing config file either locally or from the pantry path."""
import json
import os
import warnings

import s3fs
from fsspec.implementations.local import LocalFileSystem
# Duplicate code below used to handle pymongo being unavailable.
# pylint: disable-next=duplicate-code
try:
    import pymongo
except ImportError:
    _HAS_PYMONGO = False
else:
    _HAS_PYMONGO = True

from .config import get_file_system
from .pantry import Pantry
from .index.index_pandas import IndexPandas
from .index.index_sqlite import IndexSQLite
from .index.index_sql import IndexSQL


def create_pantry(**kwargs):
    """Create a weave.Pantry object using regular params or a config file.

    Three options to create a pantry:
    1: Default args, requires an Index, and pantry_path (with an optional
       file_system kwarg).
    2: A local config file which has information about the index, path, etc
    3: Only a pantry path (with an optional file_system kwarg), this will look
       for a global config file in the pantry root.

    Parameters:
    -----------
    **index: IndexABC (optional)
        The concrete implementation of an IndexABC. This is used to track
        the contents within the pantry.
    **pantry_path: str (optional)
        Path to the directory in which the pantry is located.
    **file_system: fsspec object (optional)
        The fsspec object which hosts the pantry we desire to index.
        If the file system is not passed, then the default fs is retrieved from
        the weave config.
    **config_file: str (optional)
    Optional kwargs passed to the Index constructor.

    Returns:
    -----------
    weave.Pantry object
    """
    pantry = None

    # Create a pantry in the default way using the Pantry constructor
    if all(k in kwargs for k in ["index", "pantry_path"]):
        if "file_system" in kwargs:
            file_system = kwargs.pop("file_system")
        else:
            file_system = get_file_system()
        pantry = Pantry(
            kwargs.pop("index"),
            pantry_path=kwargs.pop("pantry_path"),
            file_system=file_system,
            **kwargs,
        )
    # Create a pantry using a config file that is given to the pantry factory.
    elif "config_file" in kwargs:
        with open(
            kwargs.pop("config_file"),
            "r",
            encoding="utf-8"
        ) as config_file:
            config = json.load(config_file)

        pantry = _create_pantry_from_config(config, **kwargs)
    # Create a pantry using a config that exists inside the pantry directory.
    elif "pantry_path" in kwargs:
        if "file_system" in kwargs:
            file_system = kwargs.pop("file_system")
        else:
            file_system = get_file_system()

        try:
            with file_system.open(
                os.path.join(kwargs.pop("pantry_path"), "config.json"),
                'r'
            ) as config_file:
                config = json.load(config_file)
        except FileNotFoundError as exc:
            raise FileNotFoundError(
                "Pantry path or config file does not exist. Please add a "
                "config file path or add a config file to the pantry."
            ) from exc
        pantry = _create_pantry_from_config(config, **kwargs)
    # If pantry was unable to be created using the methods above, throw error.
    else:
        raise ValueError("Invalid kwargs passed, unable to make pantry")
    return pantry

# pylint: disable-next=too-many-branches
def _create_pantry_from_config(config, **kwargs):
    """Create the weave.pantry object from a pre-existing config file.

    Parameters:
    -----------
    config: dict
        Dictionary that contains the contents of the config.json file.
    Optional kwargs passed to the Index constructor.

    Returns:
    -----------
    weave.Pantry object
    """
    # Ensure config has the proper keys and values:
    try:
        index_type = config["index"]
        pantry_path = config["pantry_path"]
        file_system_type = config["file_system"]
    except KeyError as exc:
        raise KeyError(
            "Config file requires 'index' 'pantry_path' and "
            "'file_system' keys to build pantry from config file."
        ) from exc

    # Remove pantry kwargs to not conflict with config args.
    _ = [kwargs.pop(x, None) for x in ["index", "pantry_path", "file_system"]]

    # Parse the index type used by this pantry.
    if index_type == "IndexPandas":
        index_constructor = IndexPandas
    elif index_type == "IndexSQLite":
        index_constructor = IndexSQLite
    elif index_type == "IndexSQL":
        index_constructor = IndexSQL
    else:
        raise ValueError(f"Index Type '{index_type}' is not supported")

    # Get the file system used for this pantry.
    if file_system_type == "LocalFileSystem":
        file_system = LocalFileSystem()
    elif file_system_type == "S3FileSystem":
        if "S3_ENDPOINT" in config:
            file_system = s3fs.S3FileSystem(
                client_kwargs={"endpoint_url": config["S3_ENDPOINT"]}
            )
        else:
            # If there is no custom endpoint, use the default one in the config
            file_system = get_file_system()
    else:
        raise ValueError(f"File System Type: '{file_system_type}' is"
        f"not supported by this factory")

    # Check for optional mongo db connection config keys.
    mongo_client = None
    if "mongodb_host" in config:
        if not _HAS_PYMONGO:
            warnings.warn(
                UserWarning(
                    "Found mongo configuration keys, but pymongo is not "
                    "available on this system... Ignoring mongo config."
                )
        )
        else:
            mongo_client = pymongo.MongoClient(
                host=config["mongodb_host"],
                username=config["mongodb_username"],
                password=config["mongodb_password"],
                port=config.get("mongodb_port", 27017),
            )

    return Pantry(
        index_constructor,
        pantry_path=pantry_path,
        file_system=file_system,
        mongo_client=mongo_client,
        **kwargs,
    )



"""This script contains tests that test that pytest is deleting baskets
correctly."""
import os
import sys

# Try-Except required to make psycopg2 an optional dependency.
# Ignore pylint. This is used to explicitly show the optional dependency.
# pylint: disable=duplicate-code
try:
    import psycopg2
except ImportError:
    _HAS_PSYCOPG = False
else:
    _HAS_PSYCOPG = True
import pytest

from weave.tests.pytest_resources import PantryForTest, get_file_systems


# Create fsspec objects to be tested, and add to file_systems list.
file_systems, file_systems_ids = get_file_systems()


# Test with different fsspec file systems (above).
@pytest.fixture(
    name="set_up_tb_no_cleanup",
    params=file_systems,
    ids=file_systems_ids,
)
def fixture_set_up_tb_no_cleanup(request, tmpdir):
    """Sets up test basket fixture."""
    file_system = request.param
    temp_basket = PantryForTest(tmpdir, file_system)
    # Purposefully don't clean up pantry, it will be cleaned up in the test.
    return temp_basket


def test_weave_pytest_suffix(set_up_tb_no_cleanup):
    """Test that env var suffix works, and pantrys are still deleted."""
    # Check pantry name includes suffix if applicable.
    suffix = os.environ.get("WEAVE_PYTEST_SUFFIX", "")
    assert set_up_tb_no_cleanup.pantry_path == f"pytest-temp-pantry{suffix}"

    # Check the pantry was made.
    assert set_up_tb_no_cleanup.file_system.exists(
        set_up_tb_no_cleanup.pantry_path
    )

    # Cleanup the pantry.
    set_up_tb_no_cleanup.cleanup_pantry()

    # Check the pantry is actually deleted.
    assert not set_up_tb_no_cleanup.file_system.exists(
        set_up_tb_no_cleanup.pantry_path
    )


# Skip tests if psycopg2 is not installed.
@pytest.mark.skipif(
    "psycopg2" not in sys.modules or not _HAS_PSYCOPG
    or not os.environ.get("WEAVE_SQL_PASSWORD", False)
    or not os.environ.get("WEAVE_SQL_HOST",False),
    reason="Module 'psycopg2' required for this test "
    "AND env variables: 'WEAVE_SQL_HOST', 'WEAVE_SQL_PASSWORD'",
)
def test_github_cicd_sql_server():
    """Test that the Postgres SQL Server is properly setup in CICD."""
    # Pylint has a problem recognizing 'connect' as a valid member function
    # so we ignore that here.
    # pylint: disable-next=c-extension-no-member
    conn = psycopg2.connect(
        dbname=os.environ.get("WEAVE_SQL_DB_NAME", "weave_db"),
        host= os.environ["WEAVE_SQL_HOST"],
        user= os.environ["WEAVE_SQL_USERNAME"],
        password= os.environ["WEAVE_SQL_PASSWORD"],
        port=int(os.environ.get("WEAVE_SQL_PORT", 5432)),
        )
    cur = conn.cursor()

    # Create a temporary table for testing.
    cur.execute("""
        SELECT schema_name FROM information_schema.schemata
        WHERE schema_name = 'dbo';
    """)
    if cur.fetchall() == []:
        cur.execute("CREATE SCHEMA dbo;")
    cur.execute("""
    CREATE TABLE IF NOT EXISTS dbo.test_table (
        uuid varchar(64),
        num int
    );
    """)
    conn.commit()

    # Insert a test value, and then check we can retrieve the value.
    cur.execute("""
        INSERT INTO dbo.test_table (uuid, num) VALUES ('0001', 1);
    """)
    conn.commit()
    cur.execute("SELECT * FROM dbo.test_table")
    assert cur.fetchall() != []

    # Delete the test value, and then check it was actually deleted.
    cur.execute("""DELETE FROM dbo.test_table WHERE uuid = '0001';""")
    conn.commit()
    cur.execute("SELECT * FROM dbo.test_table;")
    assert cur.fetchall() == []

    cur.execute("""DROP TABLE dbo.test_table;""")
    conn.commit()



"""Resources for use in pytest."""

import json
import os
import sys
import io
from pathlib import Path

import pandas as pd
import s3fs
from fsspec.implementations.local import LocalFileSystem

from weave.upload import UploadBasket


def get_file_systems():
    """Returns a list of file systems and their display names."""
    file_systems = [LocalFileSystem()]
    file_systems_ids = ["LocalFileSystem"]
    if "S3_ENDPOINT" in os.environ:
        s3 = s3fs.S3FileSystem(
            client_kwargs={"endpoint_url": os.environ["S3_ENDPOINT"]}
        )
        file_systems.append(s3)
        file_systems_ids.append("S3FileSystem")
    return (file_systems, file_systems_ids)


def get_sample_basket_df():
    """Return a sample basket dataframe. THIS SHOULD ONLY BE USED AS REFERENCE
    FOR THE STRUCTURE OF THE DATAFRAME (ie column names, data types, etc.)"""
    df = pd.read_csv(io.StringIO(
        "uuid,upload_time,parent_uuids,basket_type,label,"
        "weave_version,address,storage_type\n"
        "1000,2023-10-23 16:52:43.992310+00:00,[],test_basket,test_label,"
        "1.2.0,pytest-temp-pantry/test_basket/1000,LocalFileSystem")
    )
    df["upload_time"] = pd.to_datetime(df["upload_time"])
    df["uuid"].astype(str)
    return df


def file_path_in_list(search_path, search_list):
    """Check if a file path is in a list (of file paths).

    Parameters
    ----------
    search_path: str
        The file path being searched for.
    search_list: [str]
        A list of strings (presumably file paths).

    Returns
    ----------
    bool: True if any of the file paths in the search list end with the
    file path being searched for, otherwise False.

    This allows us to determine if a file is in the list regardless of file
    system dependent prefixes such as /home/user/ which can usually be ignored.
    For example, given a list of file paths:
    ['/home/user/data/file.txt', '/home/user/data/this/is/test.txt']
    and the function is searching for 'data/file.txt',
    the function will return True as the file exists.
    """
    search_path = str(search_path)
    for file_path in search_list:
        if Path(file_path).match(search_path):
            return True

    return False


class PantryForTest:
    """Handles resources for much of weave testing."""
    def __init__(self, tmpdir, file_system, pantry_path=None):
        self.tmpdir = tmpdir
        self.file_system = file_system
        self.pantry_path = pantry_path

        if self.pantry_path is None:
            self.pantry_path = (
                "pytest-temp-pantry"
                f"{os.environ.get('WEAVE_PYTEST_SUFFIX', '')}"
            )
        self.basket_list = []
        self._set_up_pantry()

    def _set_up_pantry(self):
        """Create a temporary pantry for testing purposes."""
        try:
            self.file_system.mkdir(self.pantry_path)
        except FileExistsError:
            self.cleanup_pantry()
            self._set_up_pantry()

    def set_up_basket(
        self, tmp_dir_name, file_name="test.txt", file_content="This is a test"
    ):
        """Create a temporary (local) basket, with a single text file."""
        tmp_basket_dir = self.tmpdir.mkdir(tmp_dir_name)
        tmp_basket_txt_file = tmp_basket_dir.join(file_name)

        if file_name[file_name.rfind(".") :] == ".json":
            with open(tmp_basket_txt_file, "w", encoding="utf-8") as outfile:
                json.dump(file_content, outfile)
        else:
            tmp_basket_txt_file.write(file_content)

        return tmp_basket_dir

    def add_lower_dir_to_temp_basket(self, tmp_basket_dir):
        """Add a nested directory inside the temporary basket."""
        nested_dir = tmp_basket_dir.mkdir("nested_dir")
        nested_dir.join("another_test.txt").write("more test text")
        return tmp_basket_dir

    def upload_basket(
        self, tmp_basket_dir, uid="0000", basket_type="test_basket", **kwargs
    ):
        """Upload a temporary (local) basket to the S3 test pantry."""
        upload_items = [
            {"path": str(tmp_basket_dir.realpath()), "stub": False}
        ]

        # Ignore pylint duplicate code. Code here is required to ensure proper
        # upload, and is similar to the pantry upload call--obviously.
        # pylint: disable-next=duplicate-code
        up_dir = UploadBasket(
            upload_items=upload_items,
            basket_type=basket_type,
            file_system=self.file_system,
            pantry_path=self.pantry_path,
            unique_id=uid,
            **kwargs,
        ).get_upload_path()
        return up_dir

    def cleanup_pantry(self):
        """Delete the temporary test pantry, including any uploaded baskets."""
        self.file_system.rm(self.pantry_path, recursive=True)

# This class is to facilitate creating and deleting indices for tests.
# No other functionality required, but it is needed for pytest
# pylint: disable-next=too-few-public-methods
class IndexForTest:
    """Creates an index for testing."""
    def __init__(self, index_constructor, file_system,
                 pantry_path=None, **kwargs):
        self.file_system = file_system
        self.pantry_path = pantry_path

        if self.pantry_path is None:
            self.pantry_path = (
                "pytest-temp-pantry"
                f"{os.environ.get('WEAVE_PYTEST_SUFFIX', '')}"
            )

        # This is only used in the sqlite implementation and will not affect
        # other implmentations.
        self.db_path = f"weave-{self.pantry_path}.db"

        self.index = index_constructor(
            file_system=self.file_system,
            pantry_path=self.pantry_path,
            db_path=self.db_path,
            **kwargs,
        )

    def cleanup_index(self):
        """Clean up any artifacts created by index implementations."""
        if self.index.__class__.__name__ == "IndexSQLite":
            self.index.cur.close()
            self.index.con.close()
            # Remove SQLite db file.
            if os.path.exists(self.db_path):
                os.remove(self.db_path)

        cleanup_sql_index(self.index)

def cleanup_sql_index(index):
    """Clean up any IndexSQL"""
    if index.__class__.__name__ == "IndexSQL":
        # Drop the pantry_index (User Table) if it exists.
        index.execute_sql(f"""
            DROP TABLE IF EXISTS {index.pantry_schema}.pantry_index;
        """, commit=True)

        # Drop the parent_uuids (User Table) if it exists.
        index.execute_sql(f"""
            DROP TABLE IF EXISTS {index.pantry_schema}.parent_uuids;
        """, commit=True)

        # Drop the pantry_schema (Schema) if it exists.
        index.execute_sql(f"""
            DROP SCHEMA IF EXISTS {index.pantry_schema};
        """, commit=True)

def get_pymongo_skip_reason():
    """Move long string into shared function"""
    return ("Module: 'pymongo' required for this test AND env "
    "variables: 'MONGODB_HOST', 'MONGODB_USERNAME', 'MONGODB_PASSWORD'")

def get_pymongo_skip_condition():
    """Move long conditional into shared function"""
    return not all([
        "pymongo" in sys.modules,
        "MONGODB_HOST" in os.environ,
        "MONGODB_USERNAME" in os.environ,
        "MONGODB_PASSWORD" in os.environ,
    ])



"""
Weave
=====

Weave can be used to facilitate the creation, maintenance, and access of
coordinated, complex data storage.

Weave is used to create baskets. A basket is used to store an atomic unit (i.e.
the smallest sensible unit) of data, as well as its associated metadata,
including lineage.

--> Note that there is a Basket class available in Weave. Use it to access
individual baskets.

A collection of baskets is known as a pantry. A pantry can be accessed using
its Index.

--> Note that you can use the Index class to access a pantry's Index. You can
also upload new baskets to the pantry using Index.upload_basket().
"""

from .basket import Basket
from .index.index_pandas import IndexPandas
from .index.index_sqlite import IndexSQLite
from .index.index_sql import IndexSQL
from .pantry import Pantry
from .pantry_factory import create_pantry
from .mongo_loader import MongoLoader

__version__ = "1.13.3"

__all__ = [
    "Basket",
    "IndexPandas",
    "IndexSQLite",
    "IndexSQL",
    "Pantry",
    "MongoLoader",
    "create_pantry",
]



"""Contains functions and classes concerning the upload functionality."""

import hashlib
from importlib import metadata
import json
import math
import os
import tempfile
import uuid
from datetime import datetime, timezone as tz
from pathlib import Path
import s3fs

from fsspec.implementations.local import LocalFileSystem
from .config import get_file_system, prohibited_filenames


def validate_upload_item(upload_item, **kwargs):
    """Validates an upload_item."""
    source_file_system = kwargs.get("source_file_system", LocalFileSystem())
    if not isinstance(upload_item, dict):
        raise TypeError(
            "'upload_item' must be a dictionary: "
            f"'upload_item = {upload_item}'"
        )

    expected_schema = {"path": str, "stub": bool}
    for key, value in upload_item.items():
        if key not in expected_schema:
            raise KeyError(
                f"Invalid upload_item key: '{key}'"
                f"\nExpected keys: {list(expected_schema.keys())}"
            )
        if not isinstance(value, expected_schema[key]):
            raise TypeError(
                f"Invalid upload_item type: '{key}: {type(value)}'"
                f"\nExpected type: {expected_schema[key]}"
            )
    if not (
        source_file_system.exists(upload_item["path"])
        or os.path.exists(upload_item["path"])
    ):
        raise FileExistsError(
            f"'path' does not exist: '{upload_item['path']}'"
        )


def derive_integrity_data(file_path, byte_count=10**8, **kwargs):
    """Derives basic integrity data from a file.

    This function takes in a file path and calculates
    the file checksum, file size, and access date (current time).

    Parameters
    ----------
    file_path : str
        Path to file from which integrity data will be derived.
    byte_count: int (default=10**8)
        If the file size is greater than 3 * byte_count, the checksum
        will be calculated from the beginning, middle, and end bytes
        of the file.
        For example: If the file size is 10 bytes long
        and the byte_count is 2, the checksum will be calculated from bytes
        1, 2 (beginning two bytes), 5, 6 (middle two bytes) and 9, 10
        (last two bytes). This option is provided to speed up checksum
        calculation for large files.

    Returns
    ----------
    Dictionary
     {
      'file_size': bytes (int),
      'hash': sha256 hash (string),
      'access_date': current date/time (string),
      'source_path': path to the original source of data (string),
      'byte_count': byte count used for generated checksum (int)
     }
    """
    source_file_system = kwargs.get("source_file_system", LocalFileSystem())
    if not isinstance(file_path, str):
        raise TypeError(f"'file_path' must be a string: '{file_path}'")
    file_exists = (
        source_file_system.exists(file_path) if isinstance(
            source_file_system, s3fs.S3FileSystem
        ) else os.path.exists(file_path)
    )
    if not file_exists:
        raise FileExistsError(f"'file_path' does not exist: '{file_path}'")

    if not isinstance(byte_count, int):
        raise TypeError(f"'byte_count' must be an int: '{byte_count}'")

    if byte_count <= 0:
        raise ValueError(
            f"'byte_count' must be greater than zero: '{byte_count}'"
        )

    max_byte_count = 300 * 10**6
    if byte_count > max_byte_count:
        raise ValueError(
            f"'byte_count' must be less than or equal to {max_byte_count}"
            f" bytes: '{byte_count}'"
        )

    if isinstance(source_file_system, s3fs.S3FileSystem):
        file_size = source_file_system.du(file_path)
    else:
        file_size = os.path.getsize(file_path)

    if file_size <= byte_count * 3:
        with source_file_system.open(file_path, "rb") as file:
            sha256_hash = hashlib.sha256(file.read()).hexdigest()
    else:
        hasher = hashlib.sha256()
        midpoint = file_size / 2.0
        midpoint_seek_position = math.floor(midpoint - byte_count / 2.0)
        end_seek_position = file_size - byte_count
        with source_file_system.open(file_path, "rb") as file:
            hasher.update(file.read(byte_count))
            file.seek(midpoint_seek_position)
            hasher.update(file.read(byte_count))
            file.seek(end_seek_position)
            hasher.update(file.read(byte_count))
        sha256_hash = hasher.hexdigest()

    return {
        "file_size": file_size,
        "hash": sha256_hash,
        "access_date": datetime.now(tz.utc).isoformat(),
        "source_path": file_path,
        "byte_count": byte_count,
    }


class UploadBasket:
    """This class abstracts functionality used by upload_basket."""

    def __init__(
        self,
        upload_items,
        **kwargs,
    ):
        """Initializes the Basket_Class.

        Parameters
        ----------
        upload_items : [dict]
            List of python dictionaries with the following schema:
            {
                'path': path to the file or folder being uploaded (str),
                'stub': true/false (bool)
            }
            'path' can be a file or folder to be uploaded. Every filename and
            folder name must be unique. If 'stub' is set to True, integrity
            data will be included without uploading the actual file or folder.
            Stubs are useful when original file source information is desired
            without uploading the data itself. This is especially useful when
            dealing with large files.
        **upload_directory: str (required)
            Path where basket is to be uploaded (on the upload FS).
        **unique_id: str (required)
            Unique ID to identify the basket once uploaded.
        **basket_type: str
            Type of basket being uploaded.
        **parent_ids: [str] (optional)
            List of unique ids associated with the parent baskets
            used to derive the new basket being uploaded.
        **metadata: dict (optional)
            Python dictionary that will be written to metadata.json
            and stored in the basket in the upload FS.
        **label: str (optional)
            Optional user friendly label associated with the basket.
        **source_file_system: fsspec object (optional)
            The origin file system (ie s3fs, local fs, etc).
            If none, the local fs will be used
        **file_system: fsspec object (optional)
            The file system to upload to (ie s3fs, local fs, etc).
            If None, it will use the default fs from the weave.config.
        **pantry_path: str
            Path to the pantry that will hold this basket.
        Please note that either the upload_directory OR the basket_type must
        be provided. IT IS RECOMMENDED that the user simply provide the
        basket_type as this will allow the library to choose a good unique_id,
        and keep the pantry organized.
        """
        parent_uids = kwargs.get("parent_ids", [])
        basket_metadata = kwargs.get("metadata", {})

        # If there are no files, parent uuids are provided, and metadata is
        # provided, then it is a metadata-only basket.
        if not upload_items and (not parent_uids or not basket_metadata):
            raise ValueError(r"Files are required to upload a basket. If you "
                             r"want a metadata-only basket, please include "
                             r"metadata and parent uid(s)")

        self.upload_items = upload_items
        self.kwargs = kwargs
        # We cannot use with in this case, as we want the temp dir to persist
        # beyond this function.
        # pylint: disable-next=consider-using-with
        self.temp_dir = tempfile.TemporaryDirectory()
        self.temp_dir_path = self.temp_dir.name
        self.run_logic()

    def run_logic(self):
        """Handles running the functions that make up the class."""
        self.sanitize_args()
        self.check_that_upload_dir_does_not_exist()

        try:
            self.setup_temp_dir_for_staging_prior_to_fs()
            self.upload_files_and_stubs_to_fs()
            self.create_and_upload_basket_json_to_fs()
            self.upload_basket_metadata_to_fs()
            self.upload_basket_supplement_to_fs()

            if self.kwargs.get("test_clean_up", False):

                class TestException(Exception):
                    """Custom exception for test excepting purposes."""

                raise TestException("Test Clean Up")

        except Exception as the_exception:
            if self.fs_upload_path_exists():
                self.clean_out_fs_upload_dir()
            raise the_exception

    def sanitize_upload_basket_kwargs(self):
        """Sanitizes kwargs for upload_basket."""
        kwargs_schema = {
            "test_clean_up": bool,
            "file_system": object,
            "source_file_system": object,
            "upload_directory": str,
            "unique_id": str,
            "basket_type": str,
            "parent_ids": list,
            "metadata": dict,
            "label": str,
            "weave_version": str,
            "pantry_path": str,
            "test_prefix": str,
        }
        for key, value in self.kwargs.items():
            if key not in kwargs_schema:
                raise KeyError(f"Invalid kwargs argument: '{key}'")
            if not isinstance(value, kwargs_schema[key]):
                raise TypeError(
                    f"Invalid datatype: '{key}: "
                    f"must be type {kwargs_schema[key]}'"
                )
        # parent_ids requires further examination:
        parent_ids = self.kwargs.get("parent_ids", None)
        if (parent_ids is not None) and not (
            all(isinstance(x, str) for x in parent_ids)
        ):
            raise TypeError(
                f"'parent_ids' must be a list of strings: '{parent_ids}'"
            )
        # I think it is wise to ignore pylint here because we should only
        # set self.file_system *after* we have sanitized it.
        # pylint: disable-next=attribute-defined-outside-init
        self.file_system = self.kwargs.get("file_system", None)
        if self.file_system is None:
            # pylint: disable-next=attribute-defined-outside-init
            self.file_system = get_file_system()
        # pylint: disable-next=attribute-defined-outside-init
        self.source_file_system = self.kwargs.get(
            "source_file_system", LocalFileSystem()
        )

    def sanitize_upload_basket_non_kwargs(self):
        """Sanitize upload_basket's non kwargs args."""
        if not isinstance(self.upload_items, list):
            raise TypeError(
                "'upload_items' must be a list of dictionaries: "
                f"'{self.upload_items}'"
            )

        if not all(isinstance(x, dict) for x in self.upload_items):
            raise TypeError(
                "'upload_items' must be a list of dictionaries: "
                f"'{self.upload_items}'"
            )

        # Validate self.upload_items
        local_path_basenames = []
        for upload_item in self.upload_items:
            validate_upload_item(
                upload_item,
                file_system=self.file_system,
                source_file_system=self.source_file_system,
            )
            local_path_basename = os.path.basename(Path(upload_item["path"]))
            if local_path_basename in prohibited_filenames:
                raise ValueError(
                    f"'{local_path_basename}' " "filename not allowed"
                )
            # Check for duplicate file/folder names
            if local_path_basename in local_path_basenames:
                raise ValueError(
                    f"'upload_item' folder and file names must be unique:"
                    f" Duplicate Name = {local_path_basename}"
                )
            local_path_basenames.append(local_path_basename)

    def sanitize_args(self):
        """Sanitize all args with one call."""
        self.sanitize_upload_basket_kwargs()
        self.sanitize_upload_basket_non_kwargs()

    def _get_path(self):
        """Either make sure upload_path is in kwargs or create it."""

        if "unique_id" not in self.kwargs:
            self.kwargs["unique_id"] = uuid.uuid1().hex
        if "upload_directory" not in self.kwargs:
            if not ("pantry_path" and "basket_type" in self.kwargs):
                raise ValueError(
                    "Please provide either the 'upload_directory' or a "
                    "combination of 'pantry_path' and 'basket_type' as kwargs "
                    "for UploadBasket"
                )
            self.kwargs["upload_directory"] = os.path.join(
                self.kwargs.get("test_prefix", ""),
                self.kwargs.get("pantry_path"),
                self.kwargs.get("basket_type"),
                self.kwargs.get("unique_id"),
            )

    def check_that_upload_dir_does_not_exist(self):
        """Ensure that upload directory does not previously exist.

        This averts some errors with uploading to a directory that already
        exists.
        """

        self._get_path()
        upload_directory = self.kwargs.get("upload_directory")
        if self.file_system.isdir(upload_directory):
            raise FileExistsError(
                f"'upload_directory' already exists: '{upload_directory}''"
            )

    def setup_temp_dir_for_staging_prior_to_fs(self):
        """Sets up a temporary directory to hold stuff before upload to FS."""
        self.file_system.mkdir(self.kwargs.get("upload_directory"))

    def upload_files_and_stubs_to_fs(self):
        """Kicks off uploading files and stubs to the file_system"""
        supplement_data = {
            "integrity_data": [],
            "upload_items": self.upload_items,
        }
        for upload_item in self.upload_items:
            item_path = Path(upload_item["path"])
            if self.source_file_system.isdir(item_path):
                for root, _, files in self.source_file_system.walk(item_path):
                    for name in files:
                        local_path = os.path.join(root, name)
                        file_int_dat = self.handle_file_integrity_data(
                            local_path, upload_item, item_path
                        )
                        supplement_data["integrity_data"].append(file_int_dat)
            else:
                file_int_dat = self.handle_file_integrity_data(
                    str(item_path), upload_item, item_path
                )
                supplement_data["integrity_data"].append(file_int_dat)
        self.kwargs["supplement_data"] = supplement_data

    def handle_file_integrity_data(self, local_path, upload_item, item_path):
        """Gathers the file integrity data, handles stub logic"""
        file_int_dat = derive_integrity_data(
            str(local_path),
            file_system=self.file_system,
            source_file_system=self.source_file_system,
        )
        if upload_item["stub"] is False:
            file_int_dat["stub"] = False
            file_upload_path = self.construct_file_upload_path(
                local_path, item_path
            )
            file_int_dat["upload_path"] = str(file_upload_path)
            self.handle_file_upload(local_path, file_upload_path)
        else:
            file_int_dat["stub"] = True
            file_int_dat["upload_path"] = "stub"
        return file_int_dat

    def construct_file_upload_path(self, local_path, item_path):
        """Constructs the file_upload_path variable"""
        return os.path.join(
            self.kwargs.get("upload_directory"),
            os.path.relpath(local_path, os.path.split(item_path)[0]),
        )

    def handle_file_upload(self, local_path, file_upload_path):
        """Upload the file to fs.

        Depending on the input and output filesystems the behavior changes
        slightly."""
        base_path = os.path.split(file_upload_path)[0]
        if not self.file_system.exists(base_path):
            self.file_system.mkdir(base_path)
        if self.source_file_system == self.file_system:
            self.file_system.copy(local_path, file_upload_path)
        elif isinstance(self.source_file_system, s3fs.S3FileSystem):
            self.source_file_system.get(local_path, file_upload_path)
        else:
            self.file_system.upload(local_path, file_upload_path)

    def create_and_upload_basket_json_to_fs(self):
        """Creates and dumps a JSON containing basket metadata."""
        basket_json_path = os.path.join(
            self.temp_dir_path, "basket_manifest.json"
        )
        basket_json = {}
        basket_json["uuid"] = self.kwargs.get("unique_id")
        basket_json["upload_time"] = datetime.now(tz.utc).isoformat()
        basket_json["parent_uuids"] = self.kwargs.get("parent_ids", [])
        basket_json["basket_type"] = self.kwargs.get("basket_type")
        basket_json["label"] = self.kwargs.get("label", "")
        basket_json["weave_version"] = metadata.version("weave-db")

        with open(basket_json_path, "w", encoding="utf-8") as outfile:
            json.dump(basket_json, outfile)
        self.file_system.upload(
            basket_json_path,
            os.path.join(
                self.kwargs.get("upload_directory"), "basket_manifest.json"
            ),
        )

    def upload_basket_metadata_to_fs(self):
        """Dumps metadata to tempdir, and then uploads to FS."""
        metadata_path = os.path.join(
            self.temp_dir_path, "basket_metadata.json"
        )
        if "metadata" in self.kwargs:
            with open(metadata_path, "w", encoding="utf-8") as outfile:
                json.dump(self.kwargs.get("metadata"), outfile, default=str)
            self.file_system.upload(
                metadata_path,
                os.path.join(
                    self.kwargs.get("upload_directory"), "basket_metadata.json"
                ),
            )

    def upload_basket_supplement_to_fs(self):
        """Dumps metadata to tempdir, and then uploads to FS."""
        supplement_json_path = os.path.join(
            self.temp_dir_path, "basket_supplement.json"
        )
        with open(supplement_json_path, "w", encoding="utf-8") as outfile:
            json.dump(self.kwargs.get("supplement_data"), outfile)
        self.file_system.upload(
            supplement_json_path,
            os.path.join(
                self.kwargs.get("upload_directory"), "basket_supplement.json"
            ),
        )

    def fs_upload_path_exists(self):
        """Returns True if FS upload_path has been created, else False."""
        return self.file_system.exists(self.kwargs.get("upload_directory"))

    def clean_out_fs_upload_dir(self):
        """Removes everything from upload_path inside of FS."""
        self.file_system.rm(
            self.kwargs.get("upload_directory"), recursive=True
        )

    def get_upload_path(self):
        """Gets upload path from kwargs and returns it."""
        upload_path = self.kwargs.get("upload_directory")
        if upload_path is not None:
            return upload_path
        raise ValueError("Somehow 'upload_path' is not available in kwargs")



"""This script contains tests that test that pytest is deleting baskets
correctly."""

import os
from unittest import mock

# Try-Except required to make sqlalchemy an optional dependency.
try:
    import importlib
    assert importlib.util.find_spec('psycopg2')
    assert importlib.util.find_spec('sqlalchemy')
except ImportError:
    _HAS_REQUIRED_DEPS = False
except AssertionError:
    _HAS_REQUIRED_DEPS = False
else:
    _HAS_REQUIRED_DEPS = True
import pytest
from fsspec.implementations.local import LocalFileSystem

from weave.index.index_sql import IndexSQL
from weave.tests.pytest_resources import IndexForTest, get_file_systems
from weave.tests.pytest_resources import PantryForTest
from weave.tests.pytest_resources import get_sample_basket_df

# Create fsspec objects to be tested, and add to file_systems list.
file_systems, file_systems_ids = get_file_systems()


# Test with different fsspec file systems (above).
@pytest.fixture(
    name="test_pantry",
    params=file_systems,
    ids=file_systems_ids,
)
def fixture_test_pantry(request, tmpdir):
    """Sets up test pantry for the tests"""
    file_system = request.param
    test_pantry = PantryForTest(tmpdir, file_system)
    yield test_pantry
    test_pantry.cleanup_pantry()


@pytest.fixture(
    name="test_index",
    params=[IndexSQL],
    ids=["IndexSQL"],
)
def fixture_test_index(request):
    """Sets up test index for the tests"""
    index_constructor = request.param
    test_index = IndexForTest(index_constructor, LocalFileSystem())
    yield test_index
    test_index.cleanup_index()


# Skip tests if sqlalchemy is not installed.
@pytest.mark.skipif(
    not _HAS_REQUIRED_DEPS
    or not os.environ.get("WEAVE_SQL_PASSWORD", False),
    reason="Modules: 'psycopg2', 'sqlalchemy' required for this test "
    "AND env variables: 'WEAVE_SQL_HOST', 'WEAVE_SQL_PASSWORD'",
)
# Mock the environment variables for the test.
@mock.patch.dict(os.environ, {}, clear=True)
def test_index_sql_no_env_vars():
    """Test that the SQL Index will raise an error if the required env vars
    are not set.
    """
    with pytest.raises(KeyError) as err:
        IndexSQL(LocalFileSystem(), "weave-test-pantry")

    msg = "'The following environment variables must be set to" \
          " use this class: WEAVE_SQL_HOST, WEAVE_SQL_USERNAME, " \
          "WEAVE_SQL_PASSWORD.'"
    assert (
        str(err.value) == msg
    )


# Skip tests if sqlalchemy is not installed.
@pytest.mark.skipif(
    not _HAS_REQUIRED_DEPS
    or not os.environ.get("WEAVE_SQL_PASSWORD", False),
    reason="Modules: 'psycopg2', 'sqlalchemy' required for this test "
    "AND env variables: 'WEAVE_SQL_HOST', 'WEAVE_SQL_PASSWORD'",
)
def test_index_sql_properties_are_read_only():
    """Test that the properties of the SQL Index (database_name, pantry_schema)
    are read only, and cannot be changed during execution.
    """
    pantry_path = (
        "pytest-temp-pantry"
        f"{os.environ.get('WEAVE_PYTEST_SUFFIX', '')}"
    )

    ind = IndexSQL(LocalFileSystem(), pantry_path)

    # Disabling protected access because we want to test this variable
    original_db_name = ind._database_name # pylint: disable=W0212
    original_schema_name = pantry_path.replace("-", "_")

    with pytest.raises(AttributeError):
        ind.database_name = "new_db_name"

    with pytest.raises(AttributeError):
        ind.pantry_schema = "new_schema_name"

    assert ind.database_name == original_db_name
    assert ind.pantry_schema == original_schema_name


# Skip tests if sqlalchemy is not installed.
@pytest.mark.skipif(
    not _HAS_REQUIRED_DEPS
    or not os.environ.get("WEAVE_SQL_PASSWORD", False),
    reason="Modules: 'psycopg2', 'sqlalchemy' required for this test "
    "AND env variables: 'WEAVE_SQL_HOST', 'WEAVE_SQL_PASSWORD'",
)
def test_index_sql_tracks_different_pantries():
    """Test that the SQL Index will track different baskets using schemas."""
    sample_basket_df = get_sample_basket_df()
    uuid = str(sample_basket_df["uuid"].iloc[0])

    pantry_path = (
        "pytest-temp-pantry"
        f"{os.environ.get('WEAVE_PYTEST_SUFFIX', '')}"
    )

    ind_1 = IndexSQL(LocalFileSystem(), pantry_path + "_1")
    ind_2 = IndexSQL(LocalFileSystem(), pantry_path + "_2")

    # Perform tracks and untracks on both indices, and ensure they are correct.
    ind_1.track_basket(sample_basket_df)
    assert len(ind_1) == 1
    assert len(ind_2) == 0

    ind_2.track_basket(sample_basket_df)
    assert len(ind_1) == 1
    assert len(ind_2) == 1

    ind_1.untrack_basket(uuid)
    assert len(ind_1) == 0
    assert len(ind_2) == 1

    ind_2.untrack_basket(uuid)
    assert len(ind_1) == 0
    assert len(ind_2) == 0

    # Need to manually clean up the database in this test.
    ind_1.execute_sql(f"DROP TABLE {ind_1.pantry_schema}.pantry_index;",
        commit=True)
    ind_1.execute_sql(f"DROP TABLE {ind_1.pantry_schema}.parent_uuids;",
        commit=True)
    ind_1.execute_sql(f"DROP SCHEMA {ind_1.pantry_schema};",
        commit=True)
    ind_2.execute_sql(f"DROP TABLE {ind_2.pantry_schema}.pantry_index;",
        commit=True)
    ind_2.execute_sql(f"DROP TABLE {ind_2.pantry_schema}.parent_uuids;",
        commit=True)
    ind_2.execute_sql(f"DROP SCHEMA {ind_2.pantry_schema};",
        commit=True)


# Skip tests if sqlalchemy is not installed.
@pytest.mark.skipif(
    not _HAS_REQUIRED_DEPS
    or not os.environ.get("WEAVE_SQL_PASSWORD", False),
    reason="Modules: 'psycopg2', 'sqlalchemy' required for this test "
    "AND env variables: 'WEAVE_SQL_HOST', 'WEAVE_SQL_PASSWORD'",
)
def test_index_sql_track_basket_adds_to_parent_uuids(test_index):
    """Test that track_basket adds necessary rows to the parent_uuids table."""
    sample_basket_df = get_sample_basket_df()
    uuid = "1000"
    sample_basket_df["uuid"] = uuid

    # Add uuids to the parent_uuids of the df.
    sample_basket_df.parent_uuids = [["0001", "0002", "0003"]]

    # Track the basket.
    test_index.index.track_basket(sample_basket_df)

    ind = test_index.index
    # Manually check the parent_uuids table.
    rows, _ = ind.execute_sql(
        f"SELECT * FROM {test_index.index.pantry_schema}.parent_uuids"
    )

    # Check we have the expected values.
    assert len(rows) == 3
    assert str(rows[0]) == str(("1000", "0001"))
    assert str(rows[1]) == str(("1000", "0002"))
    assert str(rows[2]) == str(("1000", "0003"))

    # Untrack the basket and ensure values are removed from parent_uuids table.
    test_index.index.untrack_basket(uuid)
    rows, _ = ind.execute_sql(
        f"SELECT * FROM {test_index.index.pantry_schema}.parent_uuids"
    )
    assert len(rows) == 0


# Skip tests if sqlalchemy is not installed.
@pytest.mark.skipif(
    not _HAS_REQUIRED_DEPS
    or not os.environ.get("WEAVE_SQL_PASSWORD", False),
    reason="Modules: 'psycopg2', 'sqlalchemy' required for this test "
    "AND env variables: 'WEAVE_SQL_HOST', 'WEAVE_SQL_PASSWORD'",
)
def test_drop_index_deletes_sql_schema(test_index):
    """Test that drop_index drops the schema."""
    sample_basket_df = get_sample_basket_df()
    uuid = "1000"
    sample_basket_df["uuid"] = uuid

    ind = test_index.index

    # Track the basket.
    ind.track_basket(sample_basket_df)

    assert len(ind) == 1

    # Drop the index (drops the db schema)
    ind.drop_index()

    # Check the db is closed when we try to use it.
    with pytest.raises(Exception) as exc:
        len(ind)
    assert (
        f'relation "{ind.pantry_schema}.pantry_index" does not exist'
        in str(exc)
    )

    # Check the db schema is dropped.
    rows, _ = ind.execute_sql(
        "SELECT schema_name FROM information_schema.schemata;"
    )
    assert not any(ind.pantry_schema == row[0] for row in rows)

    # Recreate the db using clear_index so the test doesn't crash during
    # cleanup, as we previously dropped the schema.
    ind.clear_index()


# Skip tests if sqlalchemy is not installed.
@pytest.mark.skipif(
    not _HAS_REQUIRED_DEPS
    or not os.environ.get("WEAVE_SQL_PASSWORD", False),
    reason="Modules: 'psycopg2', 'sqlalchemy' required for this test "
    "AND env variables: 'WEAVE_SQL_HOST', 'WEAVE_SQL_PASSWORD'",
)
def test_index_sql_check_duplicate_schema_name(test_index):
    """Test that the duplicate schema error is not raised when using different
    case-sensitive paths."""
    pantry_path = test_index.index.pantry_path
    # Test that we can create a SQL Index using the same pantry path in
    # different pantry_path casing.
    _ = IndexSQL(LocalFileSystem(), pantry_path.upper())
    _ = IndexSQL(LocalFileSystem(), pantry_path.lower())



import os
from setuptools import setup

def read(rel_path: str) -> str:
    here = os.path.abspath(os.path.dirname(__file__))
    # intentionally *not* adding an encoding option to open.
    # See: https://github.com/pypa/virtualenv/issues/201#issuecomment-3145690
    with open(os.path.join(here, rel_path)) as file_path:
        return file_path.read()


def get_version(rel_path: str) -> str:
    for line in read(rel_path).splitlines():
        if line.startswith("__version__"):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    raise RuntimeError("Unable to find version string.")


with open("README.md", "r") as f:
    long_description = f.read()

setup(
    name="weave-db",
    version=get_version("weave/__init__.py"),
    description="Library to facilitate the creation and maintenance of " +
                "complex data warehouses.",
    packages=["weave", "weave/tests", "weave/index"],
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/309thEDDGE/weave",
    author="309thEDDGE",
    license="GNU General Public",
    classifiers=[
        "License :: OSI Approved :: GNU General Public License (GPL)",
        "Programming Language :: Python :: 3.10",
        "Operating System :: OS Independent",
    ],
    install_requires=["pandas", "s3fs", "fsspec", "jsonschema"],
    extras_require={
        "extras": ["pymongo", "psycopg2-binary", "sqlalchemy"],
    },
    python_requires=">=3.10",
)



"""Wherein is contained the Abstract Base Class for Index.
"""

import abc
import warnings
from datetime import datetime


class IndexABC(abc.ABC):
    """Abstract Base Class for the Index."""

    @abc.abstractmethod
    def __init__(self, file_system, pantry_path, **kwargs):
        """Initializes the Index class.

        Parameters
        ----------
        file_system: fsspec object
            The fsspec object which hosts the pantry we desire to index.
        pantry_path: str
            Path to the pantry root which we want to index.
        **setup_config: dict (required)
            Existing metadata for the Index.
        Optional kwargs controlled by concrete implementations.
        """

        self._file_system = file_system
        self._pantry_path = pantry_path
        self.setup_config = kwargs.get("setup_config", {})
        self.generate_config()

    @property
    @abc.abstractmethod
    def file_system(self):
        """The file system of the pantry referenced by this Index."""

    @property
    @abc.abstractmethod
    def pantry_path(self):
        """The pantry path referenced by this Index."""

    @abc.abstractmethod
    def generate_metadata(self, **kwargs):
        """(Deprecated) Generate the metadata for the index."""
        warnings.warn(
            UserWarning("This function is being deprecated in a future weave "
                        "release. Please use generate_config() instead.")
        )
        return self.generate_config(**kwargs)

    @abc.abstractmethod
    def generate_config(self, **kwargs):
        """Generate the setup config for the index.

        Parameters
        ----------
        Optional kwargs controlled by concrete implementations.

        Returns
        ----------
        A dictionary of metadata for the index.
        """
        self.setup_config["index_type"] = str(self)

    @abc.abstractmethod
    def generate_index(self, **kwargs):
        """Populates the index from the file system.

        Generate the index by scraping the pantry and adding the manifest data
        of found baskets to the index.

        Parameters
        ----------
        Optional kwargs controlled by concrete implementations.
        """

    @abc.abstractmethod
    def clear_index(self, refresh=False, **kwargs):
        """Deletes/clears the previously populated index.

        Delete the index entirely, optionally regenerating a new one
        if the refresh flag is set to True.

        Parameters
        ----------
        refresh: bool (default=False)
            Regenerates the index after clearing the old one.
        Optional kwargs controlled by concrete implementations.
        """

    @abc.abstractmethod
    def to_pandas_df(self, max_rows=None, offset=0, **kwargs):
        """Returns the pandas dataframe representation of the index.

        Parameters
        ----------
        max_rows: int or None (default=None)
            Max rows returned in the pandas dataframe. If None, all rows will
            be returned.
        offset: int (default=0)
            Offset from the beginning of the index to begin the query

        Optional kwargs controlled by concrete implementations.

        Returns
        ----------
        Returns a pandas dataframe of the manifest data of the baskets in the
        pantry.
        """

    @abc.abstractmethod
    def track_basket(self, entry_df, **kwargs):
        """Track a basket (or many baskets) from the pantry with the Index.

        Parameters
        ----------
        entry_df : pandas.DataFrame
            Uploaded baskets to append to the index.
        Optional kwargs controlled by concrete implementations.
        """

    @abc.abstractmethod
    def untrack_basket(self, basket_address, **kwargs):
        """Remove a basket from being tracked of given UUID or path.

        Parameters
        ----------
        basket_address: str or [str]
            Argument can take one of two forms: either a path to the basket
            directory, or the UUID of the basket. These may also be passed in
            as a list.
        Optional kwargs controlled by concrete implementations.
        """

    @abc.abstractmethod
    def get_rows(self, basket_address, **kwargs):
        """Returns a pd.DataFrame row information of given UUID or path.

        Parameters
        ----------
        basket_address: str or [str]
            Argument can take one of two forms: either a path to the basket
            directory, or the UUID of the basket. These may also be passed in
            as a list.
        Optional kwargs controlled by concrete implementations.

        Returns
        ----------
        pandas.DataFrame of the Manifest information for the requested
        basket(s).
        """

    @abc.abstractmethod
    def get_parents(self, basket_address, **kwargs):
        """Returns a pandas dataframe of all parents of a basket.

        Parameters
        ----------
        basket_address: str
            Argument can take one of two forms: either a path to the basket
            directory, or the UUID of the basket.
        Optional kwargs controlled by concrete implementations.

        Returns
        ----------
        pandas.DataFrame containing all the manifest data AND generation level
        of parents (and recursively their parents) of the given basket.
        """

    @abc.abstractmethod
    def get_children(self, basket_address, **kwargs):
        """Returns a pandas dataframe of all children of a basket.

        Parameters
        ----------
        basket_address: str
            Argument can take one of two forms: either a path to the basket
            directory, or the UUID of the basket.
        Optional kwargs controlled by concrete implementations.

        Returns
        ----------
        pandas.DataFrame containing all the manifest data AND generation level
        of children (and recursively their children) of the given basket.
        """

    @abc.abstractmethod
    def get_baskets_of_type(self, basket_type, max_rows=None,
                            offset=0, **kwargs):
        """Returns a pandas dataframe containing baskets of basket_type.

        Parameters
        ----------
        basket_type: str
            The basket type to filter for.
        max_rows: int or None (default=None)
            Max rows returned in the pandas dataframe. If None, all rows will
            be returned.
        offset: int (default=0)
            Offset from the beginning of the index to begin the query

        Optional kwargs controlled by concrete implementations.

        Returns
        ----------
        pandas.DataFrame containing the manifest data of baskets of the type.
        """

    @abc.abstractmethod
    def get_baskets_of_label(self, basket_label, max_rows=None,
                             offset=0, **kwargs):
        """Returns a pandas dataframe containing baskets with label.

        Parameters
        ----------
        basket_label: str
            The label to filter for.
        max_rows: int or None (default=None)
            Max rows returned in the pandas dataframe. If None, all rows will
            be returned.
        offset: int (default=0)
            Offset from the beginning of the index to begin the query

        Optional kwargs controlled by concrete implementations.

        Returns
        ----------
        pandas.DataFrame containing the manifest data of baskets with the
        label.
        """

    @abc.abstractmethod
    def get_baskets_by_upload_time(self, start_time=None, end_time=None,
                                   max_rows=None, offset=0, **kwargs):
        """Returns a pandas dataframe of baskets uploaded between two times.

        Parameters
        ----------
        start_time: datetime.datetime (default=None)
            The start datetime object to filter between. If None, will filter
            from the beginning of time.
        end_time: datetime.datetime (default=None)
            The end datetime object to filter between. If None, will filter
            to the current datetime.
        max_rows: int or None (default=None)
            Max rows returned in the pandas dataframe. If None, all rows will
            be returned.
        offset: int (default=0)
            Offset from the beginning of the index to begin the query.

        Optional kwargs controlled by concrete implementations.

        Returns
        ----------
        pandas.DataFrame containing the manifest data of baskets uploaded
        between the start and end times.
        """

        if start_time is not None and not isinstance(start_time,datetime):
            raise ValueError("start_time is not datetime object.")
        if end_time is not None and not isinstance(end_time,datetime):
            raise ValueError("end_time is not datetime object.")

    @abc.abstractmethod
    def query(self, expr, **kwargs):
        """Returns a pandas dataframe of the results of the query.

        Parameters
        ----------
        expr: str
            An expression passed to the backend. An example could be a SQL or
            pandas query. Largely dependent on concrete implementations.
        Optional kwargs controlled by concrete implementations.

        Returns
        ----------
        pandas.DataFrame of the resulting query.
        """

    @abc.abstractmethod
    def __len__(self):
        """Returns the number of baskets in the index."""

    @abc.abstractmethod
    def __str__(self):
        """Returns the str instantiation type of this Index (ie 'IndexSQL')."""
